import pytest
import os
from pyonir.core.media import BaseMedia, ImageFormat
from pathlib import Path

frontend_path = Path(os.path.join(os.path.dirname(__file__), "frontend", "static"))
# Create a temporary image file
@pytest.fixture
def temp_media_file(): return os.path.join(frontend_path, "test_image.png")

@pytest.fixture
def test_metadata():
    return {
        "name": "test_image",
        "width": 800,
        "height": 600,
        "created_on": "2024-01-01 00:00:00"
    }

def test_basic_file_properties(temp_media_file):
    media = BaseMedia(temp_media_file)

    assert media.file_path == temp_media_file
    assert media.file_ext == ".png"
    assert media.file_name == "test_image"
    assert media.file_dirname == str(Path(temp_media_file).parent.name)
    assert media.slug == f"{media.file_dirname}/test_image.png"

def test_encode_filename(temp_media_file, test_metadata):
    file_ext = temp_media_file.split('.')[-1]
    encoded = BaseMedia.encode_filename(temp_media_file, test_metadata)
    assert isinstance(encoded, str)
    assert encoded.endswith(f".{file_ext}")
    assert "=" not in encoded  # No padding characters
    assert "/" not in encoded  # URL-safe encoding

def test_decode_filename():
    # Test with a known encoded filename
    encoded = "bmFtZT10ZXN0X2ltYWdlJndpZHRoPTgwMCZoZWlnaHQ9NjAw"
    decoded = BaseMedia.decode_filename(encoded)

    assert isinstance(decoded, dict)
    assert decoded.get('name') == "test_image"
    assert decoded.get('width') == "800"
    assert decoded.get('height') == "600"

def test_rename_media_file(temp_media_file, test_metadata):
    media = BaseMedia(temp_media_file)
    old_path = media.file_path

    # Set test metadata
    media.data = test_metadata
    media.rename_media_file()

    assert media.file_path != old_path
    assert os.path.exists(media.file_path)
    assert not os.path.exists(old_path)
    # Set back to original for cleanup
    media.rename_media_file(os.path.basename(old_path))
    assert media.file_path == old_path

def test_resize_image():
    from PIL import Image

    # Create a real test image
    img_path = os.path.join(frontend_path, "test_resize.jpg")
    test_image = Image.new('RGB', (100, 100), color='red')
    test_image.save(img_path)

    media = BaseMedia(str(img_path))
    sizes = [(50, 50), (25, 25)]
    media.resize(sizes)

    # Check if thumbnail directory exists with resized images
    thumb_dir = Path(media.file_dirpath) / media.file_name
    assert thumb_dir.exists()

    # Check if resized files exist
    for width, height in sizes:
        resized_file = thumb_dir / f"{media.file_name}--{width}x{height}{media.file_ext}"
        assert resized_file.exists()

    # Cleanup
    os.remove(img_path)
    for f in thumb_dir.glob("*"):
        os.remove(f)
    thumb_dir.rmdir()

def test_compress_image():
    from PIL import Image

    # Create a test image
    input_path = frontend_path / "test_compress.jpg"
    output_path = frontend_path / "test_compress_output.jpg"

    test_image = Image.new('RGB', (100, 100), color='red')
    test_image.save(input_path)

    # Test compression
    BaseMedia.compress_image(str(input_path), str(output_path), quality=50)

    assert output_path.exists()
    # Compressed file should be smaller
    assert output_path.stat().st_size <= input_path.stat().st_size

    # Cleanup
    os.remove(input_path)
    os.remove(output_path)

def test_get_media_data():
    from PIL import Image

    # Create a test image
    img_path = frontend_path / "test_metadata.jpg"
    test_image = Image.new('RGB', (800, 600), color='red')
    test_image.save(img_path)

    media_data = BaseMedia.get_media_data(str(img_path))

    assert isinstance(media_data, dict)
    if media_data:
        assert 'width' in media_data
        assert 'height' in media_data
        assert media_data['width'] == 800
        assert media_data['height'] == 600

def test_image_thumbnails(temp_media_file):
    media = BaseMedia(os.path.join(frontend_path, "test_image.png"))
    media.resized_to(40, 40)
    t = media.thumbnails
    assert t.get(f'40x40') is not None




from dataclasses import is_dataclass
from datetime import datetime
from enum import EnumType
from typing import get_type_hints, Any, Tuple
from typing import get_origin, get_args, Union, Callable, Mapping, Iterable, Generator
from collections.abc import Iterable as ABCIterable, Mapping as ABCMapping, Generator as ABCGenerator

from sqlmodel import SQLModel

from pyonir.core.parser import DeserializeFile
from pyonir.core.schemas import GenericQueryModel
from pyonir.core.utils import get_attr, deserialize_datestr


def is_iterable(tp):
    origin = get_origin(tp) or tp
    return isinstance(origin, type) and issubclass(origin, ABCIterable) and not issubclass(origin, (str, bytes))

def is_generator(tp):
    origin = get_origin(tp) or tp
    return isinstance(origin, type) and issubclass(origin, ABCGenerator)

def is_mappable_type(tp):
    if tp == dict: return True
    origin = get_origin(tp)
    args = get_args(tp)
    return isinstance(origin, type) and issubclass(origin, ABCMapping)

def is_scalar_type(tp):
    sclrs = (int, float, str, bool, EnumType)
    return tp in sclrs or (isinstance(tp, type) and issubclass(tp, sclrs))

def is_custom_class(tp):
    return isinstance(tp, type) and not tp.__module__ == "builtins"

def unwrap_optional(tp):
    """Unwrap Optional[T] → T, else return tp unchanged"""
    origin_tp = get_origin(tp)
    if is_mappable_type(origin_tp):
        key_tp, value_tp = get_args(tp)
        return origin_tp, key_tp, get_args(value_tp)
    if is_iterable(origin_tp):
        value_tps = get_args(tp)
        return origin_tp, None, value_tps
    if origin_tp is Union:
        args = [unwrap_optional(a) for a in get_args(tp) if a is not type(None)]
        if len(args):
            args = args[0]
            return args[0], None, args[1]
            # res = [arg for arg, *rest in args]
    return tp, None, None

def is_callable_type(tp):
    return get_origin(tp) is Callable

def is_optional_type(tp):
    return get_origin(tp) is Union and type(None) in get_args(tp)

def is_option_type(t):
    if get_origin(t) is not Union: return t
    return [arg for arg in get_args(t) if arg is not type(None)][0]

def coerce_union(t, v):
    try:
        return t(v)
    except Exception as exc:
        print(f"failed to coerce {v} into {t}")
        return None

def coerce_unions(union_types: list[type], v: any):

    _value = None
    for utyp in union_types:
        if _value is not None: break
        try:
            _value = utyp(v)
        except Exception as exc:
            print(f"failed to coerce {v} into {utyp}")
            pass
    return _value


def collect_type_hints(t, public_only=True):
    hints = get_type_hints(t)
    try:
        init_hints = get_type_hints(t.__init__)
        hints.update(init_hints)
        if hints.get('return'):
            del hints['return']
    except Exception as exc:
        pass
    return {k:v for k,v in hints.items() if public_only and k[0]!='_'}

def required_parameters(cls):
    import inspect
    sig = inspect.signature(cls.__init__)
    required = []
    for name, param in sig.parameters.items():
        if name in ("self","args","kwargs"):  # skip self, *args, **kwargs
            continue
        if param.default is inspect.Parameter.empty:
            required.append(name)
    return required

def set_attr(target: object, attr: str, value: Any):
    if isinstance(target, dict):
        target.update({attr: value})
    else:
        setattr(target, attr, value)

is_sqlmodel_field = lambda t: callable(getattr(t,'default_factory', None))
is_sqlmodel = lambda t: isinstance(t, type) and issubclass(t, SQLModel)

def func_request_mapper(func: Callable, pyonir_request: 'BaseRequest') -> dict:
    """Map request data to function parameters"""
    from pyonir import PyonirRequest
    from pyonir import PyonirApp
    from pyonir.core import Auth
    import inspect
    # param_type_map = collect_type_hints(func)
    default_args = {}
    default_args.update(**pyonir_request.path_params.__dict__)
    default_args.update(**pyonir_request.query_params.__dict__)
    default_args.update(**pyonir_request.form)
    cls_args = {}


    sig = inspect.signature(func)
    hints = get_type_hints(func)
    params_info = {}

    for name, param in sig.parameters.items():
        param_type = hints.get(name, Any)
        param_value = default_args.get(name)
        default = (
            param.default if param.default is not inspect.Parameter.empty else None
        )
        if param_type in (PyonirApp, PyonirRequest):
            param_value = pyonir_request.app if param_type == PyonirApp else pyonir_request
        elif issubclass(param_type, Auth):
            param_value = param_type(pyonir_request, pyonir_request.app)
        else:
            param_value = cls_mapper(param_value, param_type) if param_value else default
        set_attr(cls_args, name, param_value)
        params_info[name] = {"type": param_type, "default": param_value or default}

    return cls_args

def coerce_value_to_type(value: Any, target_type: Union[type, Tuple[type]], default_factory: Callable = None) -> Any:
    """Coerce a value to the specified target type."""
    is_nullable = is_optional_type(target_type)
    actual_type, map_key_type, union_types = unwrap_optional(target_type) if not isinstance(target_type, tuple) else (None,None, target_type)
    if is_nullable and value is None:
        return None

    if union_types:
        _value = None
        if is_mappable_type(actual_type) and map_key_type:
            _value = {map_key_type(k): coerce_value_to_type(v, union_types) for k, v in value.items()}
        elif is_iterable(actual_type):
            _value = [coerce_value_to_type(v, union_types) for v in value]
        else:
            has_type = type(value) in union_types
            _value = value if has_type else coerce_unions(union_types, value)
        return _value

    if isinstance(value, actual_type):
        return value
    elif value is None and callable(default_factory):
        return default_factory()
    elif value is None and is_sqlmodel_field(default_factory):
        return default_factory.default_factory()
    elif is_scalar_type(actual_type):
        return actual_type(value)
    elif issubclass(actual_type, datetime):
        return deserialize_datestr(value)
    elif is_callable_type(actual_type) or callable(value):
        return value
    elif is_custom_class(actual_type):
        return cls_mapper(value, actual_type)
    else:
        try:
            return actual_type(value)
        except Exception as e:
            return value

def cls_mapper(file_obj: Union[dict, DeserializeFile], cls: Union['BaseSchema', type]):
    """Recursively map dict-like input into `cls` with type-safe field mapping."""
    from pyonir.core.schemas import BaseSchema

    if hasattr(cls, '__skip_parsely_deserialization__'):
        return file_obj
    is_generic = isinstance(cls, GenericQueryModel)
    is_base = issubclass(cls, BaseSchema) if not is_generic else False
    cls_ins = cls() if is_base else {}
    field_hints = cls.__fields__ if is_base or is_generic else [(k,v) for k,v in collect_type_hints(cls).items()]
    alias_keymap = cls.__alias__ if hasattr(cls, '__alias__') else {}
    is_frozen = cls.__frozen__ if hasattr(cls, '__frozen__') else False

    # normalize data source
    mapper_key = getattr(cls, '__nested_field__', None)
    data = get_attr(file_obj, mapper_key or 'data') or {}

    # assign primary fields
    processed = set()
    for name, hint in field_hints:
        if name.startswith("_") or name == "return":
            continue
        name_alias = get_attr(alias_keymap, name, None)
        # access untyped value from data, file_obj, cls (in that order)
        for ds in (data, file_obj, cls):
            value = get_attr(ds, name_alias or name)
            if value is not None: break
        if value is None:
            set_attr(cls_ins, name, value)
            continue
        # Handle containers
        custom_mapper_fn = getattr(cls, f'map_to_{name}', None)
        if custom_mapper_fn:
            value = custom_mapper_fn(value)
        else:
            fn_factory = (value if callable(value) or is_sqlmodel_field(value) else None)
            if fn_factory:
                value = None
            value = coerce_value_to_type(value, hint, default_factory=fn_factory)
        set_attr(cls_ins, name, value)
        processed.add(name)
    if is_generic:
        cls_ins.update({'file_name': get_attr(file_obj, 'file_name'),
                        'file_created_on': get_attr(file_obj, 'file_created_on')})
        return dict_to_class(cls_ins, 'GenericQueryModel')
    res = cls_ins if is_base else cls(**cls_ins)
    if not is_frozen:
        for key, value in data.items():
            if isinstance(getattr(cls, key, None), property): continue  # skip properties
            if key in processed or key[0] == '_': continue  # skip private or declared attributes
            setattr(res, key, value)
    return res if field_hints else coerce_value_to_type(file_obj, cls)

# def xcls_mapper(file_obj: object, cls: Union[type, list[type]], from_request=None):
#     """Recursively map dict-like input into `cls` with type-safe field mapping."""
#
#     from pyonir.core.schemas import BaseSchema
#
#     if hasattr(cls, '__skip_parsely_deserialization__'):
#         return file_obj
#
#     # Union types
#     if isinstance(cls, (list, tuple, set)):
#         _value = None
#         for ct in cls:
#             if isinstance(file_obj, ct):
#                 _value = ct(file_obj)
#             if _value is not None: break
#             _value = coerce_union(ct, file_obj)
#         return _value
#
#     if is_optional_type(cls) or isinstance(file_obj, cls):
#         return file_obj
#
#     # datetime passthrough
#     if cls == datetime and isinstance(file_obj, str):
#         return deserialize_datestr(file_obj)
#
#     # Scalars just wrap
#     if is_scalar_type(cls):
#         return cls(file_obj)
#     is_sqlmodel_field = lambda t: callable(getattr(value,'default_factory', None))
#     is_sqlmodel = lambda t: isinstance(t, type) and issubclass(t, SQLModel)
#     is_dclass = is_dataclass(cls) or len(required_parameters(cls)) > 0
#     is_generic_type = cls.__name__ == 'GenericQueryModel'
#     is_pyonirschema = issubclass(cls, BaseSchema)
#     param_type_map =  {k: field.annotation for k, field in cls.model_fields.items()} if is_sqlmodel(cls) else dict(cls.__fields__) if is_pyonirschema else collect_type_hints(cls)
#     data = get_attr(file_obj, 'data') or {}
#
#     # Merge nested access if ORM opts define mapper_key
#     orm_opts = getattr(cls, "_orm_options", {})
#     mapper_keys = orm_opts.get("mapper", {})
#     is_frozen = orm_opts.get("frozen")
#     access_path_to_nested = '.'.join(['data', orm_opts.get('mapper_key', cls.__name__.lower())])
#     nested_value = get_attr(file_obj, access_path_to_nested)
#     if nested_value:
#         data.update(**nested_value)
#
#     cls_args = {} if is_dclass else cls()
#     for name, hint in param_type_map.items():
#         mapper_name = get_attr(mapper_keys, name, None) #or name
#         if name.startswith("_") or name == "return":
#             continue
#
#         actual_type, *mapable = unwrap_optional(hint)
#         # value = None
#         for ds in (data, file_obj, cls): # try to get value from data, file_obj, cls (in that order)
#             value = get_attr(ds, mapper_name or name)
#             if value is not None: break
#
#         if value is None:
#             set_attr(cls_args, name, value)
#             continue
#
#         # Handle containers
#         custom_mapper_fn = getattr(cls, f'map_to_{name}', None)
#         if isinstance(value, actual_type):
#             pass
#         elif custom_mapper_fn:
#             value = custom_mapper_fn(value)
#         elif is_sqlmodel(hint):
#             value = cls_mapper(value, hint) if isinstance(value, dict) else value
#         elif is_sqlmodel_field(value):
#             value = value.default_factory()
#         elif is_scalar_type(actual_type):
#             value = actual_type(value)
#         elif is_callable_type(actual_type) or callable(value):
#             pass
#         elif is_mappable_type(actual_type) and len(mapable) and mapable[0]:
#             key_type, value_types = mapable
#             vtype = value_types[0] if len(value_types)==1 else None
#             value = {key_type(k): cls_mapper(v, vtype or value_types) for k, v in value.items()}
#         elif is_iterable(actual_type):
#             itypes = mapable[0] if len(mapable) == 1 else mapable
#             itype = itypes[0] if itypes and len(itypes) == 1 else None
#             value = [cls_mapper(v, itype) for v in value] if itype else value if isinstance(value, actual_type) else actual_type(value)
#         elif is_custom_class(actual_type):
#             value = cls_mapper(value, actual_type)
#         else:
#             try:
#                 value = actual_type(value)
#             except Exception as e:
#                 pass
#         set_attr(cls_args, name, value)
#
#     # Methods passed from request are returned to be called later
#     if callable(cls) and from_request:
#         return cls_args
#
#     res = cls(**cls_args) if is_dclass else cls_args
#     if is_generic_type:
#         data = file_obj if not data and isinstance(file_obj, dict) else data
#         keys = list(cls.__dict__.keys()) + ['file_name','file_created_on']
#         for key in keys:
#             if key[0] == '_': continue
#             _key = mapper_keys.get(key, key)
#             value = get_attr(data, _key or key) or get_attr(file_obj, _key or key)
#             set_attr(res, key, value)
#         return res
#
#     # Pass additional fields that are not specified on model
#     if not is_frozen and not is_sqlmodel(cls):
#         # keys = data.keys() if is_sqlmodel(cls) else data.model_fields.keys()
#         for key, value in data.items():
#             if isinstance(getattr(cls, key, None), property):
#                 continue  # skip properties
#             if param_type_map.get(key) or key[0] == '_':
#                 continue  # skip private or declared attributes
#             setattr(res, key, value)
#
#     return res

def dict_to_class(data: dict, name: Union[str, callable] = None, deep: bool = True) -> object:
    """
    Converts a dictionary into a class object with the given name.

    Args:
        data (dict): The dictionary to convert.
        name (str): The name of the class.
        deep (bool): If True, convert all dictionaries recursively.
    Returns:
        object: An instance of the dynamically created class with attributes from the dictionary.
    """
    # Dynamically create a new class
    cls = type(name or 'T', (object,), {}) if not callable(name) and deep!='update' else name

    # Create an instance of the class
    instance = cls() if deep!='update' else cls
    setattr(instance, 'update', lambda d: dict_to_class(d, instance, 'update') )
    # Assign dictionary keys as attributes of the instance
    for key, value in data.items():
        if isinstance(getattr(cls, key, None), property): continue
        if deep and isinstance(value, dict):
            value = dict_to_class(value, key)
        setattr(instance, key, value)

    return instance



from dataclasses import dataclass

import os, uuid
from typing import Optional, Union, List, Dict
from datetime import datetime

from pyonir import PyonirRequest, PyonirApp
from pyonir.core.schemas import BaseSchema, GenericQueryModel

from pyonir.core.mapper import cls_mapper, dict_to_class
from pyonir.core.parser import DeserializeFile
from pyonir.core.utils import parse_query_model_to_object
from pyonir.libs.plugins.navigation import Menu


# ==== Sample classes to map into ====

class Address:

    def __init__(self,
                 street: str,
                 zip_code: Optional[int] = None):
        self.street = street
        self.zip_code = zip_code

class User:

    def __init__(self,
                 uid: int,
                 name: str,
                 email: Optional[str],
                 address: Optional[Address],
                 tags: List[str],
                 meta: Dict[str, Union[str, int]]):
        self.uid = uid
        self.name = name
        self.email = email
        self.address = address
        self.tags = tags
        self.meta = meta

@dataclass
class Article:
    """Article model for testing cls_mapper with custom aliasing."""
    __alias__ = {'id': 'file_name', 'caption': 'content'}
    __frozen__ = True
    title: str
    caption: str
    alt: str
    # default factory functions called when no value provided
    id: str = uuid.uuid4
    created_on: datetime = datetime.now
    last_updated: datetime = datetime.now

article_filepath = os.path.join(os.path.dirname(__file__), 'contents', 'article.md')
# ==== Tests ====
def test_request_mapper():

    def demo_route(user_id: int, request: PyonirRequest):
        pass
    from pyonir.core.mapper import func_request_mapper
    app = PyonirApp(__file__)
    pyonir_request = PyonirRequest(None, app)
    pyonir_request.path_params = dict_to_class({'user_id': '42'})
    pyonir_request.query_params = dict_to_class({})
    args = func_request_mapper(demo_route, pyonir_request)
    pass

def test_cls_mapper_menu():
    data = {
        'menu': {
            'url': '/home',
            'slug': 'home',
            'title': 'Home',
            'group': 'main',
            'rank': 1,
            'status': 'active'
        }
    }
    menu_obj = cls_mapper(data, Menu)
    assert isinstance(menu_obj, Menu)
    assert menu_obj.url == '/home'
    assert menu_obj.slug == 'home'
    assert menu_obj.title == 'Home'
    assert menu_obj.group == 'main'
    assert menu_obj.rank == 1
    assert menu_obj.status == 'active'
    assert menu_obj.name == 'Home'

def test_parsely_to_custom_mapping():
    obj = DeserializeFile(article_filepath)
    article = cls_mapper(obj, Article)
    assert isinstance(article, Article)
    assert article.id is not None
    assert article.caption == obj.data['content']


def test_no_hint_mapping():
    generic_model = GenericQueryModel('title,url,author,date:file_created_on')
    obj = {"title": "hunter", "author": "Alice", "url": "/foo", "date": None}
    genmodel = cls_mapper(obj, generic_model)
    assert genmodel.author == "Alice"
    assert genmodel.url == '/foo'

def test_scalar_mapping():
    obj = {"uid": "123", "name": "Alice", "email": None, "address": None, "tags": [], "meta": {}}
    user = cls_mapper(obj, User)
    assert isinstance(user.uid, int)
    assert user.uid == 123
    assert user.name == "Alice"
    assert user.email is None

def test_optional_mapping():
    obj = {"id": 1, "name": "Bob", "email": "bob@test.com", "address": None, "tags": [], "meta": {}}
    user = cls_mapper(obj, User)
    assert user.email == "bob@test.com"
    obj2 = {"id": 2, "name": "Charlie", "email": None, "address": None, "tags": [], "meta": {}}
    user2 = cls_mapper(obj2, User)
    assert user2.email is None

def test_nested_object():
    obj = {
        "uid": 10, "name": "Diana", "email": "diana@test.com",
        "address": {"street": "Main St", "zip_code": "90210"},
        "tags": ["admin", "staff"],
        "meta": {"age": "30", "score": 95}
    }
    user = cls_mapper(obj, User)
    assert isinstance(user.address, Address)
    assert user.address.street == "Main St"
    assert isinstance(user.address.zip_code, int)
    assert user.address.zip_code == 90210
    assert user.meta["score"] == 95  # int conversion
    assert user.meta["age"] == "30"  # str preserved

def test_list_mapping():
    obj = {
        "uid": 20, "name": "Eva", "email": None,
        "address": None,
        "tags": ["one", "two"],
        "meta": {}
    }
    user = cls_mapper(obj, User)
    assert isinstance(user.tags, list)
    assert user.tags == ["one", "two"]

def test_dict_mapping_with_union():
    obj = {
        "uid": 30, "name": "Frank", "email": None,
        "address": None,
        "tags": [],
        "meta": {"age": 42, "nickname": "franky"}
    }
    user = cls_mapper(obj, User)
    assert isinstance(user.meta["age"], int)
    assert isinstance(user.meta["nickname"], str)



import os
import sqlite3
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional, Type, Union, Iterator, Generator

from sortedcontainers import SortedList

from pyonir.core.mapper import cls_mapper
from pyonir.core.parser import DeserializeFile
from pyonir.core.schemas import BaseSchema
from pyonir.pyonir_types import PyonirApp, AppCtx
from pyonir.core.utils import get_attr


@dataclass
class BasePagination:
    limit: int = 0
    max_count: int = 0
    curr_page: int = 0
    page_nums: list[int, int] = field(default_factory=list)
    items: list['DeserializeFile'] = field(default_factory=list)

    def __iter__(self) -> Iterator['DeserializeFile']:
        return iter(self.items)

class DatabaseService(ABC):
    """Stub implementation of DatabaseService with env-based config + builder overrides."""

    def __init__(self, app: PyonirApp, db_name: str = '') -> None:
        # Base config from environment
        from pyonir.core.utils import get_attr
        self.app = app
        self.connection: Optional[sqlite3.Connection] = None
        self._db_name: str = db_name
        self._config: object = get_attr(app.env, 'database')
        self._database: str = '' # the db address or name. path/to/directory, path/to/sqlite.db
        self._driver: str = 'sqlite' #the db context fs, sqlite, mysql, pgresql, oracle
        self._host: str = ''
        self._port: int = 0
        self._username: str = ''
        self._password: str = ''

    @property
    def datastore_path(self):
        """Path to the app datastore directory"""
        return self.app.datastore_dirpath

    @property
    def db_name(self) -> str:
        return self._db_name

    @property
    def driver(self) -> Optional[str]:
        return self._driver

    @property
    def host(self) -> Optional[str]:
        return self._host

    @property
    def port(self) -> Optional[int]:
        return self._port

    @property
    def username(self) -> Optional[str]:
        return self._username

    @property
    def password(self) -> Optional[str]:
        return self._password

    @property
    def database(self) -> Optional[str]:
        return self._database

    # --- Builder pattern overrides ---
    def set_driver(self, driver: str) -> "DatabaseService":
        self._driver = driver
        return self

    def set_database(self, database_dirpath: str = None) -> "DatabaseService":
        if self.driver.startswith('sqlite') or self.driver == 'fs':
            assert self.db_name is not None
            database_dirpath = os.path.join(database_dirpath or self.datastore_path, self.db_name)
        self._database = database_dirpath
        return self

    def set_host(self, host: str) -> "DatabaseService":
        self._host = host
        return self

    def set_port(self, port: int) -> "DatabaseService":
        self._port = port
        return self

    def set_username(self, username: str) -> "DatabaseService":
        self._username = username
        return self

    def set_password(self, password: str) -> "DatabaseService":
        self._password = password
        return self

    def set_db_name(self, name: str):
        self._db_name = name
        return self

    # --- Database operations ---
    def get_existing_columns(self, table_name: str) -> Dict[str, str]:
        cursor = self.connection.cursor()
        cursor.execute(f"PRAGMA table_info({table_name});")
        return {row[1]: row[2] for row in cursor.fetchall()}

    def get_pk(self, table: str, with_columns: bool = False):
        cursor = self.connection.cursor()
        cursor.execute(f"PRAGMA table_info('{table}')")
        pk = "id"
        columns = {}
        for col in cursor.fetchall():
            cid, name, type_, notnull, dflt_value, pk = col
            columns[name] = type_
            if pk == 1:
                pk = name
                if not with_columns: break
        columns.update({"__pk__": pk})
        return pk if not with_columns else columns

    def rename_table_columns(self, table_name: str, rename_map: dict):
        """
        Renames columns in database schema table
        :param table_name: database table
        :param rename_map: dict with key as existing column name and value as the new name
        :return:
        """
        cursor = self.connection.cursor()
        existing_cols = self.get_existing_columns(table_name)

        for old_name, new_name in rename_map.items():
            if old_name in existing_cols and new_name not in existing_cols:
                sql = f"ALTER TABLE {table_name} RENAME COLUMN {old_name} TO {new_name};"
                cursor.execute(sql)
                print(f"[RENAME] {old_name} → {new_name}")
        self.connection.commit()

    def add_table_columns(self, table_name: str, column_map: dict):
        """
        Adds new table column in database schema table
        :param table_name: database table
        :param column_map: dict with key as column name and value as the type
        :return:
        """
        cursor = self.connection.cursor()
        existing_cols = self.get_existing_columns(table_name)

        for col, dtype in column_map.items():
            if col not in existing_cols:
                sql = f"ALTER TABLE {table_name} ADD COLUMN {col} {dtype};"
                cursor.execute(sql)
                print(f"[ADD] Column '{col}' added ({dtype})")
        self.connection.commit()


    def has_table(self, table_name: str) -> bool:
        """checks if table exists"""
        if not self.connection:
            raise RuntimeError('Database service has not been initialized')
        cursor = self.connection.cursor()
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table' AND name=?", (table_name,))
        return bool(cursor.fetchone())

    @abstractmethod
    def destroy(self):
        """Destroy the database or datastore."""
        if self.driver == "sqlite" and self.database and os.path.exists(self.database):
            os.remove(self.database)
            print(f"[DEBUG] SQLite database at {self.database} has been deleted.")
        elif self.driver == "fs" and self.database and os.path.exists(self.database):
            import shutil
            shutil.rmtree(self.database)
            print(f"[DEBUG] File system datastore at {self.database} has been deleted.")
        else:
            raise ValueError(f"Cannot destroy unknown driver or non-existent database: {self.driver}:{self.database}")

    @abstractmethod
    def create_table_from_model(self, model: BaseSchema) -> 'DatabaseService':
        sql = model.generate_sql_table(self.driver)
        self.create_table(sql)
        return self

    @abstractmethod
    def create_table(self, sql_create: str) -> 'DatabaseService':
        """Create a table in the database."""
        if self.driver != "sqlite":
            raise NotImplementedError("Create operation is only implemented for SQLite in this stub.")
        if not self.connection:
            raise ValueError("Database connection is not established.")
        if self.has_table(sql_create):
            raise RuntimeError(f"Table {sql_create} already exists.")
        cursor = self.connection.cursor()
        cursor.execute(sql_create)
        return self

    @abstractmethod
    def connect(self) -> None:
        if not self.database:
            raise ValueError("Database must be set before connecting")

        if self.driver.startswith("sqlite"):
            print(f"[DEBUG] Connecting to SQLite database at {self.database}")
            self.connection = sqlite3.connect(self.database)
            self.connection.row_factory = sqlite3.Row
        elif self.driver == "fs":
            print(f"[DEBUG] Using file system path at {self.database}")
            Path(self.database).mkdir(parents=True, exist_ok=True)
        else:
            raise ValueError(f"Unknown driver: {self.driver}")

    @abstractmethod
    def disconnect(self) -> None:
        print(f"[DEBUG] Disconnecting from {self.driver}:{self.database}")
        if self.driver == "sqlite" and self.connection:
            self.connection.close()
            self.connection = None

    @abstractmethod
    def insert(self, table: str, entity: BaseSchema) -> Any:
        """Insert entity into backend."""
        import json
        data = entity if isinstance(entity, dict) else entity.to_dict()

        if self.driver == "sqlite":
            keys = ', '.join(data.keys())
            placeholders = ', '.join('?' for _ in data)
            query = f"INSERT INTO {table} ({keys}) VALUES ({placeholders})"
            values = tuple(json.dumps(v) if isinstance(v,(dict, list, tuple, set)) else v for v in data.values())
            cursor = self.connection.cursor()
            cursor.execute(query, values)
            self.connection.commit()
            return getattr(entity, entity.__primary_key__) if hasattr(entity,'__primary_key__') else cursor.lastrowid

        elif self.driver == "fs":
            # Save JSON file per record
            entity.save_to_file(entity.file_path)
            return os.path.exists(entity.file_path)

    @abstractmethod
    def find(self, table: str, filter: Dict = None) -> Any:
        results = []

        if self.driver == "sqlite":
            where_clause = ''
            params = ()
            if filter:
                where_clause = 'WHERE ' + ' AND '.join(f"{k} = ?" for k in filter)
                params = tuple(filter.values())
            query = f"SELECT * FROM {table} {where_clause}"
            cursor = self.connection.cursor()
            cursor.execute(query, params)
            results = [dict(row) for row in cursor.fetchall()]

        elif self.driver == "fs":
            pass

        return results

    @abstractmethod
    def update(self, table: str, key_value: Any, data: Dict) -> bool:
        """Update entity in backend using primary key."""
        if self.driver == "sqlite":
            if not self.connection:
                raise ValueError("Database connection is not established.")

            # Get schema class to find primary key
            schema_cls = None
            for row in self.connection.execute(f"SELECT * FROM {table} LIMIT 1"):
                schema_cls = type(table.capitalize(), (BaseSchema,), {k: None for k in dict(row).keys()})
                break

            pk_field = getattr(schema_cls, '_primary_key', 'id') if schema_cls else 'id'

            # Build UPDATE query
            set_clause = ', '.join(f"{k} = ?" for k in data.keys())
            query = f"UPDATE {table} SET {set_clause} WHERE {pk_field} = ?"
            values = list(data.values()) + [key_value]

            try:
                cursor = self.connection.cursor()
                cursor.execute(query, values)
                self.connection.commit()
                return cursor.rowcount > 0
            except sqlite3.Error as e:
                print(f"[ERROR] SQLite update failed: {e}")
                return False

        return False

class BaseFSQuery:
    """Base class for querying files and directories"""
    _cache: Dict[str, Any] = {}

    def __init__(self, query_path: str,
                app_ctx: AppCtx = None,
                model: Optional[object] = None,
                name_pattern: str = None,
                exclude_dirs: tuple = None,
                exclude_names: tuple = None,
                include_names: tuple = None,
                force_all: bool = True) -> None:

        self.query_path = query_path
        # self.sort_by: str = 'file_created_on'
        self.order_by: str = 'file_created_on' # column name to order items by
        self.order_dir: str = 'asc' # asc or desc
        self.limit: int = 0
        self.max_count: int = 0
        self.curr_page: int = 0
        self.page_nums: list[int, int] = None
        self.where_key: str = None
        self.sorted_files: SortedList = None
        self.files: Generator[DeserializeFile] = query_fs(query_path,
                              app_ctx = app_ctx,
                              model = model,
                              name_pattern = name_pattern,
                              exclude_dirs = exclude_dirs,
                              exclude_names = exclude_names,
                              include_names = include_names,
                              force_all = force_all)

    def set_params(self, params: dict):
        for k in ['limit', 'curr_page','max_count','page_nums','order_by','order_dir','where_key']:
            if k in params:
                if k in ('limit', 'curr_page', 'max_count') and params[k]:
                    params[k] = int(params[k])
                setattr(self, k, params[k])
        return self

    def sorting_key(self, x: any):
        if self.order_dir not in ("asc", "desc"):
            raise ValueError("order_dir must be 'asc' or 'desc'")

        def _invert(val):
            # For numbers and timestamps
            if isinstance(val, (int, float)):
                return -val
            # For strings: reverse lexicographic order
            if isinstance(val, str):
                return "".join(chr(255 - ord(c)) for c in val)
            # Fallback
            return val

        value = get_attr(x, self.order_by)

        # If sorting by datetime-like values
        if isinstance(value, datetime):
            value = value.timestamp()

        # If value is None, push it to the end consistently
        if value is None:
            return float("inf") if self.order_dir == "asc" else float("-inf")

        return value if self.order_dir == "asc" else _invert(value)

    def paginated_collection(self)-> Optional[BasePagination]:
        """Paginates a list into smaller segments based on curr_pg and display limit"""
        from sortedcontainers import SortedList

        if self.order_by:
            self.sorted_files = SortedList(self.files, self.sorting_key)
        if self.where_key:
            where_key = [self.parse_params(ex) for ex in self.where_key.split(',')]
            self.sorted_files = SortedList(self.where(**where_key[0]), self.sorting_key)
        force_all = not self.limit

        self.max_count = len(self.sorted_files)
        page_num = 0 if force_all else int(self.curr_page)
        start = (page_num * self.limit) - self.limit
        end = (self.limit * page_num)
        pg = (self.max_count // self.limit) + (self.max_count % self.limit > 0) if self.limit > 0 else 0
        pag_data = self.paginate(start=start, end=end, reverse=True) if not force_all else self.sorted_files

        return BasePagination(
            curr_page = page_num,
            page_nums = [n for n in range(1, pg + 1)] if pg else None,
            limit = self.limit,
            max_count = self.max_count,
            items = list(pag_data)
        )

    def paginate(self, start: int, end: int, reverse: bool = False):
        """Returns a slice of the items list"""
        sl = self.sorted_files.islice(start, end, reverse=reverse) if end else self.sorted_files
        return sl

    @staticmethod
    def prev_next(input_file: 'DeserializeFile'):
        """Returns the previous and next files relative to the input file"""
        from pyonir.core.mapper import dict_to_class
        prv = None
        nxt = None
        pc = BaseFSQuery(input_file.file_dirpath)
        _collection = iter(pc.files)
        for cfile in _collection:
            if cfile.file_status == 'hidden': continue
            if cfile.file_path == input_file.file_path:
                nxt = next(_collection, None)
                break
            else:
                prv = cfile
        return dict_to_class({"next": nxt, "prev": prv})

    def find(self, value: any, from_attr: str = 'file_name'):
        """Returns the first item where attr == value"""
        return next((item for item in self.sorted_files if getattr(item, from_attr, None) == value), None)

    def where(self, attr, op="=", value=None):
        """Returns a list of items where attr == value"""
        from pyonir.core.utils import get_attr

        def match(item):
            actual = get_attr(item, attr)
            if not hasattr(item, attr):
                return False
            if actual and not value:
                return True # checking only if item has an attribute
            elif op == "=":
                return actual == value
            elif op == "in" or op == "contains":
                return actual in value if actual is not None else False
            elif op == ">":
                return actual > value
            elif op == "<":
                return actual < value
            elif op == ">=":
                return actual >= value
            elif op == "<=":
                return actual <= value
            elif op == "!=":
                return actual != value
            return False
        if callable(attr): match = attr
        if not self.sorted_files:
            self.sorted_files = SortedList(self.files, lambda x: get_attr(x, self.order_by) or x)
        return filter(match, list(self.sorted_files or self.files))

    def __len__(self):
        return self.sorted_files and len(self.sorted_files) or 0

    def __iter__(self):
        return iter(self.sorted_files)

    @staticmethod
    def parse_params(param: str):
        k, _, v = param.partition(':')
        op = '='
        is_eq = lambda x: x[1]=='='
        if v.startswith('>'):
            eqs = is_eq(v)
            op = '>=' if eqs else '>'
            v = v[1:] if not eqs else v[2:]
        elif v.startswith('<'):
            eqs = is_eq(v)
            op = '<=' if eqs else '<'
            v = v[1:] if not eqs else v[2:]
            pass
        elif v[0]=='=':
            v = v[1:]
        else:
            pass
        return {"attr": k.strip(), "op":op, "value":BaseFSQuery.coerce_bool(v)}

    @staticmethod
    def coerce_bool(value: str):
        d = ['false', 'true']
        try:
            i = d.index(value.lower().strip())
            return True if i else False
        except ValueError as e:
            return value.strip()


def query_fs(abs_dirpath: str,
                app_ctx: AppCtx = None,
                model: Union[object, str] = None,
                name_pattern: str = None,
                exclude_dirs: tuple = None,
                exclude_names: tuple = None,
                include_names: tuple = None,
                force_all: bool = True) -> Generator:
    """Returns a generator of files from a directory path"""
    from pathlib import Path
    from pyonir.core.page import BasePage
    from pyonir.core.parser import DeserializeFile
    from pyonir.core.media import BaseMedia

    # results = []
    hidden_file_prefixes = ('.', '_', '<', '>', '(', ')', '$', '!', '._')
    allowed_content_extensions = ('prs', 'md', 'json', 'yaml')
    def get_datatype(filepath) -> Union[object, BasePage, BaseMedia]:
        if model == 'path': return str(filepath)
        if model == BaseMedia: return BaseMedia(filepath)
        pf = DeserializeFile(str(filepath), app_ctx=app_ctx)
        if model == 'file':
            return pf
        pf.schema = BasePage if (pf.is_page and not model) else model
        res = cls_mapper(pf, pf.schema) if pf.schema else pf
        return res

    def skip_file(file_path: Path) -> bool:
        """Checks if the file should be skipped based on exclude_dirs and exclude_file"""
        is_private_file = file_path.name.startswith(hidden_file_prefixes)
        is_excluded_file = exclude_names and file_path.name in exclude_names
        is_included_file = include_names and file_path.name in include_names
        is_allowed_file = file_path.suffix[1:] in allowed_content_extensions
        if is_included_file: return False
        if not is_private_file and force_all: return False
        return is_excluded_file or is_private_file or not is_allowed_file

    for path in Path(abs_dirpath).rglob(name_pattern or "*"):
        if skip_file(path) or path.is_dir(): continue
        yield get_datatype(path)


from abc import ABC
from typing import Optional, get_type_hints

from pyonir import PyonirApp

class BaseService(ABC):
    """
    Abstract base class defining a generic service interface for Pyonir applications.
    """
    app: Optional[PyonirApp]
    """Pyonir application instance"""

    name: str
    """Name of the service"""

    version: str
    """Version of the service"""

    endpoint: str
    """API endpoint for the service"""

    @property
    def endpoint_url(self) -> str:
        """Construct the full URL for the service endpoint."""
        return f"{self.endpoint}/{self.version}" if self.version else self.endpoint

    def generate_api(self, namespace: str = '') -> None:
        """Generate API resolvers for the service."""
        if not self.app:
            raise ValueError("Pyonir application instance is not available.")
        if self.app.server.is_active: return
        import os
        base_path = os.path.join(self.app.contents_dirpath, self.app.API_DIRNAME)
        self.app.generate_resolvers(self, base_path, namespace=namespace)

    def __init_subclass__(cls, **kwargs):
        """Validate that required attributes are defined in subclasses."""
        super().__init_subclass__(**kwargs)

        required_attrs = get_type_hints(cls)

        for attr_name, expected_type in required_attrs.items():
            if attr_name == 'app':
                continue
            if not hasattr(cls, attr_name):
                raise TypeError(
                    f"{cls.__name__} must define class attribute '{attr_name}'"
                )

            attr_value = getattr(cls, attr_name)
            if not isinstance(attr_value, expected_type):
                raise TypeError(
                    f"{cls.__name__}.{attr_name} must be of type {expected_type.__name__}, "
                    f"got {type(attr_value).__name__}"
                )

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        # SERVICE_INSTANCES.append(instance)
        return instance



import uuid
from datetime import datetime
from typing import Type, TypeVar, Any, Optional


T = TypeVar("T")

class BaseSchema:
    """
    Interface for immutable dataclass models with CRUD and session support.
    """
    __table_name__ = str()
    __fields__ = set()
    __alias__ = dict()
    __primary_key__ = str()
    __frozen__ = bool()
    _sql_create_table: Optional[str] = None
    _errors: list[dict[str, Any]]

    def __init_subclass__(cls, **kwargs):
        from pyonir.core.mapper import collect_type_hints
        table_name = kwargs.get("table_name")
        primary_key = kwargs.get("primary_key")
        dialect = kwargs.get("dialect")
        alias = kwargs.get("alias_map", {})
        frozen = kwargs.get("frozen", False)
        if table_name:
            setattr(cls, "__table_name__", table_name)
        print(f'init_subclass for {cls.__name__}')
        model_fields = set((name, typ)  for name, typ in collect_type_hints(cls).items())
        setattr(cls, "__fields__", model_fields)
        setattr(cls, "__primary_key__", primary_key or "id")
        setattr(cls, "_errors", [])
        setattr(cls, "__alias__", alias)
        setattr(cls, "__frozen__", frozen)
        cls.generate_sql_table(dialect)

    def __init__(self, **data):
        from pyonir.core.mapper import coerce_value_to_type

        for field_name, field_type in self.__fields__:
            value = data.get(field_name)
            custom_mapper_fn = getattr(self, f'map_to_{field_name}', None)
            type_factory = getattr(self, field_name, custom_mapper_fn)
            value = coerce_value_to_type(value, field_type, default_factory=type_factory) if value or type_factory else None
            setattr(self, field_name, value)

    def is_valid(self) -> bool:
        """Returns True if there are no validation errors."""
        return not self._errors

    def validate_fields(self, field_name: str = None):
        """
        Validates fields by calling `validate_<fieldname>()` if defined.
        Clears previous errors on every call.
        """
        if field_name is not None:
            validator_fn = getattr(self, f"validate_{field_name}", None)
            if callable(validator_fn):
                validator_fn()
            return
        for name in self.__dict__.keys():
            if name.startswith("_"):
                continue
            validator_fn = getattr(self, f"validate_{name}", None)
            if callable(validator_fn):
                validator_fn()

    def model_post_init(self, __context):
        """sqlmodel post init callback"""
        object.__setattr__(self, "_errors", [])
        self.validate_fields()

    def __post_init__(self):
        """Dataclass post init callback"""
        self._errors = []
        self.validate_fields()

    def save_to_file(self, file_path: str) -> bool:
        """Saves the user data to a file in JSON format"""
        from pyonir.core.utils import create_file
        return create_file(file_path, self.to_dict(obfuscate=False))

    def save_to_session(self, request: 'PyonirRequest', key: str = None, value: any = None) -> None:
        """Convert instance to a serializable dict."""
        request.server_request.session[key or self.__class__.__name__.lower()] = value

    def to_dict(self, obfuscate = True):
        """Dictionary representing the instance"""

        obfuscated = lambda attr: obfuscate and hasattr(self,'_private_keys') and attr in (self._private_keys or [])
        is_ignored = lambda attr: attr in ('file_path','file_dirpath') or attr.startswith("_") or callable(getattr(self, attr)) or obfuscated(attr)
        def process_value(key, value):
            if hasattr(value, 'to_dict'):
                return value.to_dict(obfuscate=obfuscate)
            if isinstance(value, property):
                return getattr(self, key)
            if isinstance(value, (tuple, list, set)):
                return [process_value(key, v) for v in value]
            return value
        fields = self.__fields__
        return {key: process_value(key, getattr(self, key)) for key, ktype in fields if not is_ignored(key) and not obfuscated(key)}

    def to_json(self, obfuscate = True) -> str:
        """Returns a JSON serializable dictionary"""
        import json
        return json.dumps(self.to_dict(obfuscate))

    @classmethod
    def from_file(cls: Type[T], file_path: str, app_ctx=None) -> T:
        """Create an instance from a file path."""
        from pyonir.core.parser import DeserializeFile
        from pyonir.core.mapper import cls_mapper
        prsfile = DeserializeFile(file_path, app_ctx=app_ctx)
        return cls_mapper(prsfile, cls)

    @classmethod
    def generate_sql_table(cls, dialect: str = None) -> str:
        """Generate the CREATE TABLE SQL string for this model."""
        from sqlalchemy.schema import CreateTable
        from sqlalchemy.dialects import sqlite
        from sqlalchemy.dialects import postgresql
        from sqlalchemy.dialects import mysql
        from sqlalchemy import Boolean, Float, JSON, Table, Column, Integer, String, MetaData
        dialect = dialect or "sqlite"
        PY_TO_SQLA = {
            int: Integer,
            str: String,
            float: Float,
            bool: Boolean,
            dict: JSON,
            list: JSON,
        }
        primary_key = getattr(cls, "__primary_key__", None)
        table_name = getattr(cls, '__table_name__', None) or cls.__name__.lower()
        columns = []
        has_pk = False
        for name, typ in cls.__annotations__.items():
            col_type = PY_TO_SQLA.get(typ, String)
            is_pk = name == 'id' or name == primary_key and not has_pk
            kwargs = {"primary_key": is_pk}
            columns.append(Column(name, col_type, **kwargs))
            if is_pk:
                has_pk = True
        if not has_pk:
            # Ensure at least one primary key
            columns.insert(0, Column("id", Integer, primary_key=True, autoincrement=True))
        table = Table(table_name, MetaData(), *columns)

        # Pick dialect
        if dialect == "sqlite":
            dialect_obj = sqlite.dialect()
        elif dialect == "postgresql":
            dialect_obj = postgresql.dialect()
        elif dialect == "mysql":
            dialect_obj = mysql.dialect()
        else:
            raise ValueError(f"Unsupported dialect: {dialect}")

        cls._sql_create_table = str(CreateTable(table, if_not_exists=True).compile(dialect=dialect_obj))
        return cls._sql_create_table

    @staticmethod
    def generate_date(date_value: str = None) -> datetime:
        from pyonir.core.utils import deserialize_datestr
        return deserialize_datestr(date_value or datetime.now())

    @classmethod
    def generate_id(cls) -> str:
        return uuid.uuid4().hex


class GenericQueryModel:
    """A generic model to hold dynamic fields from query strings."""
    file_created_on: str
    file_name: str
    def __init__(self, model_str: str):
        aliases = {}
        fields = set()
        for k in model_str.split(','):
            if ':' in k:
                k,_, src = k.partition(':')
                aliases[k] = src
            fields.add((k, str))

        setattr(self, "__fields__", fields)
        setattr(self, "__alias__", aliases)



# -*- coding: utf-8 -*-
import os, sys
from pyonir.core.app import BaseApp, Optional
from pyonir.core.utils import get_version
from pyonir.core.app import BaseApp, BasePlugin
from pyonir.core.database import BaseFSQuery
from pyonir.core.schemas import BaseSchema
from pyonir.core.server import BaseRequest, BaseServer

# Pyonir settings
PYONIR_DIRPATH = os.path.abspath(os.path.dirname(__file__))
PYONIR_TOML_FILE = os.path.join(os.path.dirname(PYONIR_DIRPATH), "pyproject.toml")
PYONIR_LIBS_DIRPATH = os.path.join(PYONIR_DIRPATH, "libs")
PYONIR_PLUGINS_DIRPATH = os.path.join(PYONIR_LIBS_DIRPATH, 'plugins')
PYONIR_SETUPS_DIRPATH = os.path.join(PYONIR_LIBS_DIRPATH, 'app_setup')
PYONIR_JINJA_DIRPATH = os.path.join(PYONIR_LIBS_DIRPATH, 'jinja')
PYONIR_JINJA_TEMPLATES_DIRPATH = os.path.join(PYONIR_JINJA_DIRPATH, "templates")
PYONIR_JINJA_EXTS_DIRPATH = os.path.join(PYONIR_JINJA_DIRPATH, "extensions")
PYONIR_JINJA_FILTERS_DIRPATH = os.path.join(PYONIR_JINJA_DIRPATH, "filters")

__version__: str = get_version(PYONIR_TOML_FILE)
Site: Optional[BaseApp] = None


class PyonirApp(BaseApp):pass
class PyonirServer(BaseServer): pass
class PyonirRequest(BaseRequest): pass
class PyonirCollection(BaseFSQuery): pass
class PyonirSchema(BaseSchema): pass
class PyonirPlugin(BasePlugin): pass

class Pyonir(PyonirApp):
    """Pyonir Application"""
    def __init__(self, entry_file_path: str,
                 use_themes: bool = None,
                 salt: str = None):
        """Initializes existing Pyonir application"""
        global Site
        sys.path.insert(0, os.path.dirname(os.path.dirname(entry_file_path)))
        super().__init__(entry_file_path,
                         use_themes=use_themes,
                         salt=salt)
        Site = self
        self.process_configs()
        if use_themes:
            self.configure_themes()

# def init(entry_file_path: str, use_themes: bool = None):
#     """Initializes existing Pyonir application"""
#     global Site
#     # Set Global Site instance
#     # if options: options = PyonirOptions(**(options or {}))
#     sys.path.insert(0, os.path.dirname(os.path.dirname(entry_file_path)))
#     Site = PyonirApp(entry_file_path, use_themes=use_themes)
#     Site.process_configs()
#     if use_themes:
#         Site.configure_themes()
#     return Site


import os, re, json
from typing import Tuple, Dict, List, Any, Optional, Union

from pyonir.core.utils import get_file_created, open_file, get_attr

# Pre-compile regular expressions for better performance
_RE_LEADING_SPACES = re.compile(r'^\s+')

# Constants
DICT_DELIM = ": "
LST_DLM = ":-"
LST_DICT_DLM = "-"
STR_DLM = ":` "
ILN_DCT_DLM = ":: "
BLOCK_DELIM = ":|"
BLOCK_PREFIX_STR = "==="
BLOCK_CODE_FENCE = "````"
SINGLE_LN_COMMENT = '#'
MULTI_LN_COMMENT = '#|'
LOOKUP_EMBED_PREFIX = '$'
LOOKUP_DIR_PREFIX = '$dir'
LOOKUP_DATA_PREFIX = '$data'
FILTER_KEY = '@filter'

# Global cache
EmbeddedTypes: Dict[str, Any] = {}


class FileStatuses(str):
    UNKNOWN = "unknown"
    """Read only by the system often used for temporary and unknown files"""

    PROTECTED = "protected"
    """Requires authentication and authorization. can be READ and WRITE."""

    FORBIDDEN = "forbidden"
    """System only access. READ ONLY"""

    PUBLIC = "public"
    """Access external and internal with READ and WRITE."""


class DeserializeFile:
    """Context for parsely file processing"""
    _virtual_route_filename = '.virtual_routes.md'
    _routes_dirname = "pages"
    """Directory name that contains page files served as file based routing"""

    def __init__(self,
                file_path: str,
                app_ctx: "AppCtx" = None,
                model: object = None,
                text_string: str = None):
        name, ext = os.path.splitext(os.path.basename(file_path))
        self.app_ctx = app_ctx
        self._blob_keys = []
        self.schema = model
        self.file_ext = ext
        self.file_name = name
        self.file_path = str(file_path)
        self.file_dirpath = os.path.dirname(file_path)  # path to files contents directory
        self.file_dirname = os.path.basename(self.file_dirpath)
        self.file_contents_dirpath = None
        # file data processing
        self.text_string = text_string
        self.file_lines = None
        self.file_line_count = None
        self.data: Dict = {}

        # Page specific attributes
        if not self.text_string:
            ctx_name, ctx_url, contents_dirpath, ssg_path, datastore_path = app_ctx or ("", "", "", "", "",)
            contents_relpath = (
                file_path.replace(contents_dirpath, "").lstrip("/")
                if contents_dirpath
                else ""
            )
            contents_dirname = contents_relpath.split("/")[0]
            is_page = contents_dirname == self._routes_dirname
            self.file_contents_dirpath = contents_dirpath or self.file_dirpath
            self.is_page = is_page
            self.is_home = (
                is_page and contents_relpath == f"{self._routes_dirname}/index"
            )
            self.is_virtual_route = self.file_path.endswith(
                self._virtual_route_filename
            )
            # page attributes
            if not self.is_virtual_route:
                surl = (
                    re.sub(rf"\b{contents_dirname}/\b|\bindex\b", "", contents_relpath)
                    if is_page
                    else contents_relpath
                )
                slug = (
                    f"{ctx_url or ''}/{surl}".lstrip("/")
                    .rstrip("/")
                    .lower()
                    .replace(self.file_ext, "")
                )
                url = "/" if self.is_home else "/" + slug
                self.data["url"] = url
                self.data["slug"] = slug

        # process file data
        self.deserializer()
        # Post-processing
        self.apply_filters()

    def apply_filters(self):
        """Applies filter methods to data attributes"""
        from pyonir import Site

        if not bool(self.data):
            return
        filters = self.data.get(FILTER_KEY)
        if not filters or not Site:
            return
        for filtr, datakeys in filters.items():
            for key in datakeys:
                mod_val = self.process_site_filter(
                    filtr, get_attr(self.data, key), {"page": self.data}
                )
                update_nested(key, self.data, data_update=mod_val)
        del self.data[FILTER_KEY]

    def deserializer(self):
        """Deserialize file line strings into map object"""
        if self.file_ext == ".md" or self.text_string:
            lines = open_file(self.file_path) or self.text_string
            self.file_lines = lines.strip().split("\n") if lines else []
            self.file_line_count = len(self.file_lines)
            if self.file_line_count > 0:
                # from pyonir.core.parsely import process_lines
                process_lines(self.file_lines, cursor=0, data_container=self.data, file_ctx=self)
        elif self.file_ext == ".json":
            self.data = open_file(self.file_path, rtn_as="json") or {}
        return True

    def __lt__(self, other: 'DeserializeFile') -> bool:
        """Compares two DeserializeFile instances based on their created_on attribute."""
        if not isinstance(other, DeserializeFile):
            return True
        return self.file_created_on < other.file_created_on

    @property
    def file_exists(self):
        return os.path.exists(self.file_path)

    @property
    def file_modified_on(self):  # Datetime
        from datetime import datetime
        import pytz

        return (
            datetime.fromtimestamp(os.path.getmtime(self.file_path), tz=pytz.UTC)
            if self.file_exists
            else None
        )

    @property
    def file_created_on(self):  # Datetime
        return get_file_created(self.file_path) if self.file_exists else None

    @property
    def file_status(self) -> str:  # String
        return (
            FileStatuses.PROTECTED
            if self.file_name.startswith("_")
            else FileStatuses.FORBIDDEN
            if self.file_name.startswith(".")
            else FileStatuses.PUBLIC
        )

    @staticmethod
    def process_site_filter(filter_name: str, value: any, kwargs=None):
        from pyonir import Site

        if not Site or (not Site.SSG_IN_PROGRESS and not Site.server.is_active):
            return value
        site_filter = Site.Parsely_Filters.get(filter_name)
        return site_filter(value, kwargs)

    @classmethod
    def load(cls, json_str: str) -> dict:
        """converts parsely string to python dictionary object"""
        f = cls("", text_string=json_str)
        return f.data

    @staticmethod
    def loads(data: dict) -> str:
        """converts python dictionary object to parsely string"""
        return serializer(data)

    def refresh_data(self):
        """Parses file and update data values"""
        self.data = {}
        self._blob_keys.clear()
        self.deserializer()
        self.apply_filters()

    def prev_next(self):
        from pyonir.core.database import BaseFSQuery

        if self.file_dirname != "pages" or self.is_home:
            return None
        return BaseFSQuery.prev_next(self)

    def to_named_tuple(self):
        """Returns a tuple representation of the file data"""
        from collections import namedtuple

        file_keys = [
            *self.data.keys(),
            "file_name",
            "file_ext",
            "file_path",
            "file_dirpath",
            "file_dirname",
        ]
        PageTuple = namedtuple("PageTuple", file_keys)
        return PageTuple(
            **self.data,
            file_name=self.file_name,
            file_ext=self.file_ext,
            file_path=self.file_path,
            file_dirpath=self.file_dirpath,
            file_dirname=self.file_dirname,
        )

    def output_html(self, req: "PyonirRequest") -> str:
        """Renders and html output"""
        from pyonir import Site
        from pyonir.core.page import BasePage
        from pyonir.core.mapper import cls_mapper


        # from pyonir.core.mapper import add_props_to_object
        # refresh_model = get_attr(req, 'query_params.rmodel')
        page = cls_mapper(self, self.schema or BasePage)
        Site.apply_globals({"prevNext": self.prev_next, "page": page})
        html = Site.TemplateEnvironment.get_template(page.template).render()
        Site.TemplateEnvironment.block_pull_cache.clear()
        return html

    def output_json(self, data_value: any = None, as_str=True) -> str:
        """Outputs a json string"""
        from .utils import json_serial

        data = data_value or self
        if not as_str:
            return data
        return json.dumps(data, default=json_serial)

    def generate_static_file(self, page_request=None, rtn_results=False):
        """Generate target file as html or json. Takes html or json content to save"""
        from pyonir import Site

        count = 0
        html_data = None
        json_data = None
        ctx_static_path = (
            self.app_ctx[3] if self.app_ctx and len(self.app_ctx) > 3 else ""
        )
        slug = self.data.get("slug")

        def render_save():
            # -- Render Content --
            html_data = self.output_html(page_request)
            json_data = self.output_json(as_str=False)
            # -- Save contents --
            create_file(path_to_static_html, html_data)
            create_file(path_to_static_api, json_data)
            return 2

        # -- Get static paths --
        path_to_static_api = os.path.join(
            ctx_static_path, Site.API_DIRNAME, slug, "index.json"
        )
        path_to_static_html = os.path.join(ctx_static_path, slug, "index.html")

        count += render_save()

        if page_request:
            for pgnum in range(1, page_request.paginate):
                path_to_static_html = os.path.join(
                    self.file_ssg_html_dirpath, str(pgnum + 1), "index.html"
                )
                path_to_static_api = os.path.join(
                    self.file_ssg_api_dirpath, str(pgnum + 1), "index.json"
                )
                page_request.query_params["pg"] = pgnum + 1
                count += render_save()

        # -- Return contents without saving --
        if rtn_results:
            return html_data, json_data

        return count

def parse_markdown(content, kwargs):
    """Parse markdown string using mistletoe with htmlattributesrenderer"""
    import html, mistletoe

    # from mistletoe.html_attributes_renderer import HTMLAttributesRenderer
    if not content:
        return content
    res = mistletoe.markdown(content)
    # res = mistletoe.markdown(content, renderer=HTMLAttributesRenderer)
    return html.unescape(res)


def count_tabs(str_value: str, tab_width: int = 4) -> int:
    """Returns number of tabs for provided string using cached regex"""
    try:
        match = _RE_LEADING_SPACES.match(str_value.replace('\n', ''))
        return round(len(match.group()) / tab_width) if match else 0
    except Exception:
        return 0

def update_nested(attr_path, data_src: dict, data_merge=None, data_update=None, find=None) -> tuple[bool, dict]:
    """
    Finds or updates target value based on an attribute path.

    Args:
        attr_path (list): Attribute path as list or dot-separated string.
        data_src (dict): Source data to search or update.
        data_merge (Any, optional): Value to merge.
        data_update (Any, optional): Value to replace at path.
        find (bool, optional): If True, only retrieve the value.

    Returns:
        tuple[bool, Any]: (completed, updated data or found value)
    """

    def update_value(target, val):
        """Mutates target with val depending on type compatibility."""
        if isinstance(target, list):
            if isinstance(val, list):
                target.extend(val)
            else:
                target.append(val)
        elif isinstance(target, dict) and isinstance(val, dict):
            target.update(val)
        elif isinstance(target, str) and isinstance(val, str):
            return val
        return target

    # Normalize attribute path
    if isinstance(attr_path, str):
        attr_path = attr_path.strip().split('.')
    if not attr_path:
        return True, update_value(data_src, data_merge)

    completed = len(attr_path) == 1

    # Handle list source at top-level
    if isinstance(data_src, list):
        _, merged_val = update_nested(attr_path, {}, data_merge)
        return update_nested(None, data_src, merged_val)

    # Navigate deeper if not at last key
    if not completed:
        current_data = {}
        for i, key in enumerate(attr_path):
            if find:
                current_data = (data_src.get(key) if not current_data else current_data.get(key))
            else:
                completed, current_data = update_nested(
                    attr_path[i + 1:],
                    data_src.get(key, current_data),
                    find=find,
                    data_merge=data_merge,
                    data_update=data_update
                )
                update_value(data_src, {key: current_data})
                if completed:
                    break
    else:
        # Last key operations
        key = attr_path[-1].strip()

        if find:
            return True, data_src.get(key)

        if data_update is not None:
            return completed, update_value(data_src, {key: data_update})

        # If key not in dict, wrap merge value in a dict
        if isinstance(data_src, dict) and data_src.get(key) is None:
            data_merge = {key: data_merge}

        if isinstance(data_merge, (str, int, float, bool)):
            data_src[key] = data_merge
        elif isinstance(data_src, dict):
            update_value(data_src.get(key, data_src), data_merge)
        else:
            update_value(data_src, data_merge)

    return completed, (data_src if not find else current_data)

def serializer(json_map: dict, namespace: list = [], inline_mode: bool = False, filter_params=None) -> str:
    """Converts python dictionary into parsely string"""

    if filter_params is None:
        filter_params = {}
    mode = 'INLINE' if inline_mode else 'NESTED'
    lines = []
    multi_line_keys = []
    is_block_str = False

    def pair_map(key, val, tabs):
        is_multiline = isinstance(val, str) and len(val.split("\n")) > 2
        if is_multiline or key in filter_params.get('_blob_keys', []):
            multi_line_keys.append((f"==={key.replace('content', '')}{filter_params.get(key, '')}", val.strip()))
            return
        if mode == 'INLINE':
            ns = ".".join(namespace)
            value = f"{ns}.{key}: {val}" if bool(namespace) else f"{key}: {val.strip()}"
            lines.append(value)
        else:
            if key:
                lines.append(f"{tabs}{key}: {val}")
            else:
                lines.append(f"{tabs}{val}")

    if isinstance(json_map, (str, bool, int, float)):
        tabs = '    ' * len(namespace)
        return f"{tabs}{json_map}"

    for k, val in json_map.items():
        tab_count = len(namespace) if namespace is not None else 0
        tabs = '    ' * tab_count
        if isinstance(val, (str, int, bool, float)):
            pair_map(k, val, tabs)

        elif isinstance(val, (dict, list)):
            delim = ':' if isinstance(val, dict) else ':-'
            if len(namespace) > 0:
                namespace = namespace + [k]
            else:
                namespace = [k]

            if mode == 'INLINE' and isinstance(val, list):
                ns = ".".join(namespace)
                lines.append(f"{ns}{delim}")
            elif mode == 'NESTED':
                lines.append(f"{tabs}{k}{delim}")

            if isinstance(val, dict):
                nested_value = serializer(json_map=val, namespace=namespace, inline_mode=inline_mode)
                lines.append(f"{nested_value}")
            else:
                maxl = len(val) - 1
                has_scalar = any([isinstance(it, (str, int, float, bool)) for it in val])
                for i, item in enumerate(val):
                    list_value = serializer(json_map=item, namespace=namespace, inline_mode=False)
                    lines.append(f"{list_value}")
                    if i < maxl and not has_scalar:
                        lines.append(f"    -")
            namespace.pop()

    if multi_line_keys:
        [lines.append(f"{mlk}\n{mlv}") for mlk, mlv in multi_line_keys]
    return "\n".join(lines)

def process_lookups(value_str: str, file_ctx: DeserializeFile = None) -> Optional[Union[str, dict, list]]:
    """Process $dir and $data lookups in the value string"""
    app_ctx: list = file_ctx.app_ctx
    file_contents_dirpath: str = file_ctx.file_contents_dirpath
    file_name: str = file_ctx.file_name

    def parse_ref_to_files(filepath, as_dir=0):
        from pyonir.core.database import BaseFSQuery, DeserializeFile
        from pyonir.core.utils import get_attr, import_module, parse_query_model_to_object
        from pyonir.core.schemas import GenericQueryModel

        if as_dir:
            # use proper app context for path reference outside of scope is always the root level
            # Ref parameters with model will return a generic model to represent the data value
            model = None
            generic_model_properties = query_params.get('model')
            return_all_files = query_params.get('limit','') == '*'
            if generic_model_properties:
                if '.' in generic_model_properties:
                    pkg, mod = os.path.splitext(generic_model_properties)
                    mod = mod[1:]
                    model = import_module(pkg, callable_name=mod)
                if not model:
                    model = GenericQueryModel(generic_model_properties)
                    # model = parse_query_model_to_object(generic_model_properties)
            collection = BaseFSQuery(filepath, app_ctx=app_ctx,
                                  model=model,
                                  exclude_names=(file_name, 'index.md'),
                                  force_all=return_all_files)
            data = collection.set_params(query_params).paginated_collection()
        else:
            rtn_key = has_attr_path or 'data'
            p = DeserializeFile(filepath, app_ctx=app_ctx)
            data = get_attr(p, rtn_key) or p
        return data

    raw_value = value_str.strip()
    value_str = value_str.strip()
    has_lookup = value_str.startswith((LOOKUP_DIR_PREFIX, LOOKUP_DATA_PREFIX))

    if has_lookup:
        from pyonir.core.utils import parse_url_params
        base_path = app_ctx[-1:][0] if value_str.startswith(LOOKUP_DATA_PREFIX) else file_contents_dirpath
        _query_params = value_str.split("?").pop() if "?" in value_str else False
        query_params = parse_url_params(_query_params) if _query_params else ''
        has_attr_path = value_str.split("#")[-1] if "#" in value_str else ''
        value_str = value_str.replace(f"{LOOKUP_DIR_PREFIX}/", "") \
            .replace(f"{LOOKUP_DATA_PREFIX}/", "") \
            .replace(f"?{_query_params}", "") \
            .replace(f'#{has_attr_path}', '')

        value_str = value_str.replace('../', '').replace('/*', '')
        lookup_fpath = os.path.join(base_path, *value_str.split("/"))
        if not os.path.exists(lookup_fpath):
            print({
                'ISSUE': f'FileNotFound while processing {raw_value}',
                'SOLUTION': f'Make sure the `{lookup_fpath}` file exists. Note that only valid md and json files can be processed.'
            })
            return None
        return parse_ref_to_files(lookup_fpath, os.path.isdir(lookup_fpath))
    return value_str

def deserialize_line(line_value: str, container_type: Any = None, file_ctx: DeserializeFile = None) -> Any:
    """Deserialize string value to appropriate object type"""

    if not isinstance(line_value, str):
        return line_value

    def is_num(valstr):
        valstr = valstr.strip().replace(',', '')
        if valstr.isdigit():
            return int(valstr)
        try:
            return float(valstr)
        except ValueError:
            return 'NAN'

    line_value = line_value.strip()
    has_inline_dict_expression = DICT_DELIM in line_value and ', ' not in line_value

    if has_inline_dict_expression:
        v = parse_line(line_value)
        return group_tuples_to_objects([v], parent_container=dict())

    if EmbeddedTypes.get(line_value):
        return EmbeddedTypes.get(line_value)
    is_num = is_num(line_value)
    if is_num != 'NAN':
        return is_num
    if line_value.lower() == "false":
        return False
    elif line_value.lower() == "true":
        return True
    elif isinstance(container_type, list):
        return [deserialize_line(v, file_ctx=file_ctx)  for v in line_value.split(', ')]
    elif line_value.startswith((LOOKUP_DIR_PREFIX, LOOKUP_DATA_PREFIX)):
        # if file_ctx and file_ctx.is_virtual_route:
        #     return line_value
        if '{' in line_value:
            line_value = file_ctx.process_site_filter('pyformat', line_value, file_ctx.__dict__)
        return process_lookups(line_value, file_ctx=file_ctx)
    elif line_value.startswith('$'):
        line_value = file_ctx.process_site_filter("pyformat", line_value[1:], file_ctx.__dict__)
    return line_value.lstrip('$')

def get_container_type(delim):
    if LST_DLM == delim:
        return list()
    elif DICT_DELIM == delim or DICT_DELIM.strip() == delim:
        return dict()
    else:
        return str()

def parse_line(line: str, from_block_str: bool = False, file_ctx: Any = None) -> tuple:
    """partition key value pairs"""

    try:
        start_fence_block = line.startswith((BLOCK_CODE_FENCE, BLOCK_PREFIX_STR))
        is_end_fence = from_block_str and (start_fence_block or line.strip().endswith((BLOCK_CODE_FENCE, BLOCK_DELIM)))
        if is_end_fence:
            return count_tabs(line), None, None, None, None
        iln_delim = None
        if not from_block_str:
            if line.endswith(DICT_DELIM.strip()): # normalize dict delim
                line = line[:-1] + DICT_DELIM
            iln_delim = [x for x in (
                (line.find(BLOCK_DELIM), BLOCK_DELIM),
                (line.find(STR_DLM), STR_DLM),
                (line.find(LST_DLM), LST_DLM),
                (line.find(DICT_DELIM), DICT_DELIM),
            ) if x[0] != -1]
        key, delim, value = line.partition(iln_delim[0][1]) if iln_delim else (None, None, line)
        line_type = get_container_type(delim) if delim else str()
        is_parent = not value and key is not None
        is_str_block = is_parent and isinstance(line_type, str)
        if start_fence_block:
            line = line.replace(BLOCK_CODE_FENCE, '').replace(BLOCK_PREFIX_STR, '')
            fence_key, *alias_key = line.split(' ', 1)
            key = alias_key[0] if alias_key else fence_key or 'content'
            value = None
            is_str_block = True
            is_parent = True
        if not from_block_str:
            key = key.strip() if key else None
            value = deserialize_line(value, container_type=line_type, file_ctx=file_ctx) if value else None
        elif value:
            value += '\n'
        return count_tabs(line), key, line_type, value or None, (is_str_block, is_parent)
    except Exception as e:
        raise e
        # return None, None, line.strip(), None, None

def group_tuples_to_objects(items: list[tuple],
                            parent_container: any = None,
                            use_grouped: bool = False,
                            file_ctx: DeserializeFile = None,
                            compress_strings: bool = False) -> list[dict]:
    """Groups list of tuples into list of objects or other container types """

    grouped = []
    current = {}
    is_str = isinstance(parent_container, str) and not use_grouped
    is_list = isinstance(parent_container, list) and not use_grouped
    is_dict = isinstance(parent_container, dict) and not use_grouped
    for tab_count, key, data_type, value, is_string_block in items:
        if is_str:
            parent_container += value.strip() if compress_strings else value or ''
            continue
        elif is_list:
            v = deserialize_line(value, file_ctx=file_ctx)
            value = {key: v} if isinstance(data_type, dict) else v
            parent_container.append(value)
            continue
        elif is_dict:
            update_nested(key, data_src=parent_container, data_merge=deserialize_line(value, file_ctx=file_ctx))
            continue
        if value == LST_DICT_DLM:  # separator → start a new object
            if current:
                grouped.append(current)
                current = {}
            continue

        # Normalize value for nested lists (e.g. child elements)
        if isinstance(value, list) and all(isinstance(v, tuple) for v in value):
            value = group_tuples_to_objects(value, parent_container=data_type)

        current[key] = deserialize_line(value, file_ctx=file_ctx)

    # append last object if not empty
    if current:
        grouped.append(current)

    return grouped or parent_container

def is_comment(line: str) -> bool:
    """Fast comment check using tuple unpacking"""
    return line.startswith((SINGLE_LN_COMMENT, MULTI_LN_COMMENT))

def collect_block_lines(lines: list, curr_tabs: int, is_str_block: tuple[bool, bool] = None, parent_container: Any = None, file_ctx: DeserializeFile = None) -> Tuple[list, int]:
    """Collects lines until stop string is found"""
    collected_lines = []
    cursor = 0
    is_list_dict = False
    pis_str_block, pis_parent = is_str_block or (False, False)
    is_virtual = file_ctx and file_ctx.is_virtual_route
    max = len(lines)
    while cursor < max:
        ln = lines[cursor]
        lt, lk, ld, lv, lb = parse_line(ln, from_block_str=pis_str_block, file_ctx=file_ctx)
        if lb is None:
            if cursor == max - 1:
                cursor += 1
            break
        lis_block_str, lis_parent = lb
        is_nested = lt > curr_tabs
        end_data_block = not is_nested and not pis_str_block
        end_nested_str_block = (curr_tabs > 0 and pis_str_block) and not is_nested
        if end_nested_str_block or end_data_block: break

        if not is_list_dict:
            is_list_dict = lv==LST_DICT_DLM and not ld
        if lis_parent:
            lv, _curs = collect_block_lines(lines[cursor+1:], curr_tabs=lt, is_str_block=lb, parent_container=ld, file_ctx=file_ctx)
            cursor = cursor + _curs
        cursor += 1
        collected_lines.append((lt, lk, ld, lv, lb))

    # Finalize block collection
    compress_strings = curr_tabs > 0 and pis_str_block
    collected_lines = group_tuples_to_objects(collected_lines,
                                                  use_grouped=is_list_dict,
                                                  parent_container=parent_container,
                                                  file_ctx=file_ctx,
                                                  compress_strings=compress_strings)
    return collected_lines, cursor

def process_lines(file_lines: list[str], cursor: int = 0, data_container: Dict[str, Any] = None, file_ctx: DeserializeFile = None) -> Dict[str, Any]:
    """Process lines iteratively instead of recursively"""

    if data_container is None:
        data_container = {}

    line_count = len(file_lines)
    while cursor < line_count:
        line = file_lines[cursor]

        if not line:
            cursor += 1
            continue

        if is_comment(line):
            cursor += 1
            continue

        line_tabs, line_key, line_type, line_value, is_str_block = parse_line(line, file_ctx=file_ctx)
        if line_value is None:
            line_value, _cursor = collect_block_lines(
                file_lines[cursor+1:],
                curr_tabs=line_tabs,
                is_str_block=is_str_block,
                parent_container=line_type,
                file_ctx=file_ctx
            )
            cursor = (_cursor + cursor) + 1 #if line_tabs else _cursor
        else:
            cursor += 1

        if not line_tabs:
            update_nested(line_key, data_container, data_merge=line_value)

        # Cache the $ check
        if line_key and line_key[0] == '$':
            EmbeddedTypes[line_key] = line_value
    return data_container


from dataclasses import dataclass
from datetime import datetime
import os, pytz
from typing import Any

IMG_FILENAME_DELIM = '::'  # delimits the file name and description

class PageStatus(str):
    UNKNOWN = 'unknown'
    """Read only by the system often used for temporary and unknown files"""

    PROTECTED = 'protected'
    """Requires authentication and authorization. can be READ and WRITE."""

    FORBIDDEN = 'forbidden'
    """System only access. READ ONLY"""

    PUBLIC = 'public'
    """Access external and internal with READ and WRITE."""


class BaseFile:
    """Represents a single file on file system"""

    def __init__(self, path: str = None, contents_dirpath: str = None):
        self.file_path = str(path or __file__)
        self.file_ext = os.path.splitext(self.file_path)[1]

    @property
    def file_name(self):
        name, ext = os.path.splitext(os.path.basename(self.file_path))
        self.file_ext = ext
        return name

    @property
    def file_dirpath(self):
        return os.path.dirname(self.file_path)

    @property
    def file_dirname(self):
        return os.path.basename(os.path.dirname(self.file_path))

    @property
    def file_status(self) -> str:  # String
        return PageStatus.PROTECTED if self.file_name.startswith('_') else \
            PageStatus.FORBIDDEN if self.file_name.startswith('.') else PageStatus.PUBLIC

    @property
    def created_on(self):  # Datetime
        return datetime.fromtimestamp(os.path.getctime(self.file_path), tz=pytz.UTC)

    @property
    def modified_on(self):  # Datetime
        return datetime.fromtimestamp(os.path.getmtime(self.file_path), tz=pytz.UTC)

    def to_dict(self) -> dict[str, Any]:
        return self.__dict__

@dataclass
class BasePage:
    """Represents a single page returned from a web request"""
    __alias__ = {'created_on': 'file_created_on', 'modified_on': 'file_modified_on'}
    template: str = 'pages.html'
    created_on: datetime = None
    modified_on: datetime = None

    def __lt__(self, other) -> bool:
        """Compares two BasePage instances based on their created_on attribute."""
        if not isinstance(other, BasePage):
            return True
        return self.created_on < other.created_on



from __future__ import annotations
import os
from dataclasses import dataclass, field
from typing import Optional, Union, Tuple, Callable, get_type_hints, OrderedDict, Any, List

from pyonir.core.utils import get_attr
from starlette.applications import Starlette
from starlette.requests import Request as StarletteRequest

from pyonir.core.app import BaseApp
from pyonir.core.parser import DeserializeFile
from pyonir.pyonir_types import PyonirRouters, PyonirHooks, PyonirRoute
from pyonir.core.utils import dict_to_class

TEXT_RES: str = 'text/html'
JSON_RES: str = 'application/json'
EVENT_RES: str = 'text/event-stream'

# Environments
LOCAL_ENV:str = 'LOCAL'
DEV_ENV:str = 'DEV'
PROD_ENV:str = 'PROD'




class BaseRequest:
    PAGINATE_LIMIT: int = 6

    def __init__(self, server_request: Optional[StarletteRequest], app: BaseApp):
        from pyonir.core.utils import get_attr
        from pyonir.core.auth import Auth

        self.server_response = None
        self.file: Optional[DeserializeFile] = None
        self.server_request: StarletteRequest = server_request
        self.raw_path = "/".join(str(self.server_request.url).split(str(self.server_request.base_url))) if server_request else ''
        self.method = self.server_request.method if server_request else 'GET'
        self.path = self.server_request.url.path if server_request else '/'
        self.path_params = dict_to_class(self.server_request.path_params,'path_params') if server_request else {}
        self.url = f"{self.path}" if server_request else {}
        self.slug = self.path.lstrip('/').rstrip('/')
        self.query_params = self.get_params(self.server_request.url.query) if server_request else {}
        self.parts = self.slug.split('/') if self.slug else []
        self.limit = get_attr(self.query_params, 'limit', self.PAGINATE_LIMIT)
        self.model = get_attr(self.query_params, 'model')
        self.is_home = (self.slug == '')
        self.is_api = self.parts and self.parts[0] == app.API_DIRNAME
        self.is_static = bool(list(os.path.splitext(self.path)).pop()) if server_request else False
        self.form = {}
        self.files = []
        # self.ip = self.server_request.client.host if server_request else ''
        self.host = str(self.server_request.base_url).rstrip('/') if server_request else app.host
        self.protocol = self.server_request.scope.get('type') + "://" if server_request else app.protocol
        self.headers = self.process_header(self.server_request.headers) if server_request else {}
        self.browser = self.headers.get('user-agent', '').split('/').pop(0) if self.headers else "UnknownAgent"
        if self.slug.startswith('api'): self.headers['accept'] = JSON_RES
        self.type: Union[TEXT_RES, JSON_RES, EVENT_RES] = self.headers.get('accept')
        self.status_code: int = 200
        self._app_ctx_ref = None
        self.app = app
        self.auth: Optional[Auth] = None
        self.flashes: dict = self.get_flash_messages() if server_request else {}
        if server_request:
            self.server_request.session['previous_url'] = self.headers.get('referer', '')

    @property
    def app_ctx_ref(self):
        return self._app_ctx_ref or self.app

    @property
    def session_token(self):
        """Returns active csrf token for user session"""
        if self.server_request and self.server_request.session:
            return self.server_request.session.get('csrf_token')

    @property
    def previous_url(self) -> str:
        return self.server_request.session.pop('previous_url') or '/'

    @property
    def redirect_to(self):
        """Returns the redirect URL from the request form data"""
        file_redirect = self.file.data.get('redirect_to', self.file.data.get('redirect'))
        return self.form.get('redirect_to', self.form.get('redirect', file_redirect))

    def ssg_request(self, page, params):
        self.file = page
        self.url = page.data.get('url')
        self.path = page.data.get('url')
        self.slug = page.data.get('slug')
        self.path_params.update(params)

    def redirect(self, url: str):
        """Sets the redirect URL in the request form data"""
        self.form['redirect_to'] = url

    def get_flash_messages(self) -> dict:
        """Pops and returns all flash messages from session"""
        if self.server_request and self.server_request.session:
            session_data = self.server_request.session
            flashes = session_data.pop('__flash__') if session_data.get('__flash__') else {}
            return flashes
        return {}

    def pull_flash(self, key):
        return self.flashes.get(key)

    def add_flash(self, key: str, value: any):
        flash_obj = self.server_request.session.get('__flash__') or {}
        flash_obj[key] = value
        self.server_request.session['__flash__'] = flash_obj

    def from_session(self, session_key: str) -> any:
        """Returns data from the session"""
        return self.server_request.session.get(session_key, None)

    async def process_request_data(self):
        """Get form data and file upload contents from request"""

        from pyonir import Site
        from pyonir.core.utils import expand_dotted_keys
        import json

        def secure_upload_filename(filename):
            import re
            # Strip leading and trailing whitespace from the filename
            filename = filename.strip()

            # Replace spaces with underscores
            filename = filename.replace(' ', '_')

            # Remove any remaining unsafe characters using a regular expression
            # Allow only alphanumeric characters, underscores, hyphens, dots, and slashes
            filename = re.sub(r'[^a-zA-Z0-9_.-]', '', filename)

            # Ensure the filename doesn't contain multiple consecutive dots (.) or start with one
            filename = re.sub(r'\.+', '.', filename).lstrip('.')

            # Return the filename as lowercase for consistency
            return filename.lower()

        try:
            try:
                ajson = await self.server_request.json()
                if isinstance(ajson, str): ajson = json.loads(ajson)
                self.form.update(ajson)
            except Exception as ee:
                # multipart/form-data
                form = await self.server_request.form()
                files = []
                for name, content in form.multi_items():
                    if hasattr(content, "filename"):
                        setattr(content, "ext", os.path.splitext(content.filename)[1])
                        self.files.append(content)
                    else:
                        if self.form.get(name): # convert form name into a list
                            currvalue = self.form[name]
                            if isinstance(currvalue, list):
                                currvalue.append(content)
                            else:
                                self.form[name] = [currvalue, content]
                        else:
                            self.form[name] = content
        except Exception as e:
            raise
        self.form = expand_dotted_keys(self.form, return_as_dict=True)

    def derive_status_code(self, is_router_method: bool):
        """Create status code for web request based on a file's availability, status_code property"""
        from pyonir.core.parser import FileStatuses

        code = 404
        if self.file.is_virtual_route:
            # If the file is a router method, we assume it is valid
            code = 200
        elif self.file.file_status in (FileStatuses.PROTECTED, FileStatuses.FORBIDDEN):
            self.file.data = {'template': '40x.html', 'content': f'Unauthorized access to this resource.', 'url': self.url, 'slug': self.slug}
            code = 401
        elif self.file.file_status == FileStatuses.PUBLIC or is_router_method:
            code = 200
        self.status_code = code #200 if self.file.file_exists or is_router_method else 404

    def set_app_context(self) -> None:
        """Gets the routing context from web request"""
        path_str = self.path.replace(self.app.API_ROUTE, '')
        for plg in self.app.activated_plugins:
            if not hasattr(plg, 'endpoint'): continue
            if path_str.startswith(plg.endpoint):
                self._app_ctx_ref = plg
                print(f"Request has switched to {plg.name} context")
                break

    def render_error(self):
        """Data output for an unknown file path for a web request"""
        return {
            "url": self.url,
            "method": self.method,
            "status": self.status_code,
            "res": self.server_response,
            "title": f"{self.path} was not found!",
            "content": f"Perhaps this page once lived but has now been archived or permanently removed from {self.app_ctx_ref.name}."
        }

    def fetch_virtual_route(self, app: BaseApp, url: str) -> Optional[DeserializeFile]:
        """Fetches a virtual route matching the request path for the current app context"""
        virtual_page_path = os.path.join(app.pages_dirpath, '.virtual_route') + BaseApp.EXTENSIONS['file']
        if not os.path.exists(virtual_page_path): return None
        virtual_route = DeserializeFile(virtual_page_path, app_ctx=app.app_ctx)
        vkey, vdata, vparams = self.app.server.get_virtual(url, virtual_data=virtual_route.data)
        if vparams and vkey:
            self.path_params.update(vparams)
            virtual_route.refresh_data()
            vdata = virtual_route.data.get(vkey) if vkey else vdata
        virtual_route.data = {'url': self.url, 'slug': self.slug, **(vdata or {})}
        virtual_route.is_virtual_route = bool(vkey)
        return virtual_route

    def resolve_request_to_file(self, app: BaseApp, path_str: str = None) -> DeserializeFile:
        """
        Resolve a request URL to its corresponding resource.

        The function checks plugin-provided paths first, then falls back to the main
        application's file system. If no matching file or virtual route is found,
        a 404 page is returned.
        """
        from pyonir.core.app import BaseApp
        from pyonir.core.parser import DeserializeFile
        path_str = path_str or self.path
        is_home = path_str == '/'
        ctx_route, ctx_paths = app.request_paths or ('', [])
        ctx_route = ctx_route or ''
        ctx_slug = ctx_route[1:]
        path_slug = path_str[1:]
        app_scope, *path_segments = path_slug.split('/')
        is_api_request = (len(path_segments) and path_segments[0] == app.API_DIRNAME) or path_str.startswith(app.API_ROUTE)

        # Normalize API prefix and path segments
        if is_api_request:
            path_str = path_str.replace(app.API_ROUTE, '')

        virtual_route = self.fetch_virtual_route(app, path_str)
        request_segments = [
            segment for segment in path_slug.split('/')
            if segment and segment not in (app.API_DIRNAME, ctx_slug)
        ]

        # Skip if no paths or route doesn't match
        if not ctx_paths or (not is_home and not path_str.startswith(ctx_route)):
            return None

        # Try resolving to actual file paths
        protected_segment = [s if i > len(request_segments)-1 else f'_{s}' for i,s in enumerate(request_segments)]

        for root_path in ctx_paths:
            if not is_api_request and root_path.endswith(app.API_DIRNAME): continue
            category_index = os.path.join(root_path, *request_segments, 'index.md')
            single_page = os.path.join(root_path, *request_segments) + BaseApp.EXTENSIONS['file']
            single_protected_page = os.path.join(root_path, *protected_segment) + BaseApp.EXTENSIONS['file']

            for candidate in (category_index, single_page, single_protected_page):
                if os.path.exists(candidate):
                    route_page = DeserializeFile(candidate, app_ctx=app.app_ctx)
                    if virtual_route:
                        route_page.data.update(virtual_route.data)
                        route_page.apply_filters()
                    return route_page

        errorpage = DeserializeFile('404_ERROR')
        errorpage.data = self.render_error()
        return virtual_route if virtual_route and virtual_route.is_virtual_route else errorpage

    @staticmethod
    def process_header(headers):
        nheaders = dict(headers)
        nheaders['accept'] = nheaders.get('accept', TEXT_RES).split(',', 1)[0]
        agent = nheaders.get('user-agent', '')
        nheaders['user-agent'] = agent.split(' ').pop().split('/', 1)[0]
        return nheaders

    @staticmethod
    def get_params(url, as_dict=False):
        from pyonir.core.mapper import dict_to_class
        from pyonir.core.utils import parse_url_params
        args = parse_url_params(url)
        if args.get('model'): del args['model']
        return args if as_dict else dict_to_class(args, 'query_params')

    def process_resolver(self, request: 'BaseRequest') -> Optional[Union[callable, Any]]:
        """Updates request data a callable method to execute during request."""
        from pyonir import Site
        from pyonir.core.utils import get_attr
        resolver_obj = self.file.data.get('@resolvers', {})
        resolver_action = resolver_obj.get(request.method)
        if not resolver_action: return
        resolver_path = resolver_action.get('call')
        del resolver_action['call']
        self.file.data.update(resolver_action)

        app_plugin = list(filter(lambda p: p.name == resolver_path.split('.')[0], Site.activated_plugins))
        app_plugin = app_plugin[0] if len(app_plugin) else Site
        resolver = app_plugin.reload_resolver(resolver_path)

        request.type = get_attr(resolver_action, 'headers.accept') or request.type
        request.form.update(resolver_action)
        if not resolver:
            request.auth.response = request.auth.responses.ERROR.response(message=f"Unable to resolve endpoint")
        return resolver

class BaseServer(Starlette):

    def __init__(self, app: BaseApp):
        self.is_active = False
        self.app = app
        self.ws_routes = []
        self.sse_routes = []
        self.auth_routes = []
        self.endpoints = set()
        self.url_map = {} # named reference to routes by function name
        self.route_map = {} # named reference to routes by path
        self.resolvers = {}
        self.services = {}
        self.request: Optional[BaseRequest] = None
        self.virtual_routes: Optional[OrderedDict] = None

        self.initialize_starlette()


    @property
    def allowed_hosts(self):
        """Returns a list of domains allowed to access the web application"""
        return ['localhost', '*.localhost']

    def run_uvicorn_server(self, uvicorn_options: dict = None):
        """Starts the webserver"""
        import uvicorn, sys
        from pathlib import Path

        # """Uvicorn web server configurations"""
        # Uvicorn’s config only allows one binding method at a time:
        # TCP socket → use host + port (+ optional SSL)
        # Unix domain socket → use uds (+ optional SSL)
        uvicorn_options = uvicorn_options or {}
        if not uvicorn_options:
            if self.app.is_dev:
                uvicorn_options.update({
                    "port": self.app.port,
                    "host": self.app.host,
                })
            else:
                uvicorn_options = {'uds': self.app.unix_socket_filepath}

            if self.app.is_secure:
                uvicorn_options["ssl_keyfile"] = self.app.ssl_key_file
                uvicorn_options["ssl_certfile"] = self.app.ssl_cert_file

        # Setup logs
        Path(self.app.logs_dirpath).mkdir(parents=True, exist_ok=True)
        # Initialize routers
        self.init_pyonir_endpoints(self.app)
        print(f"/************** ASGI APP SERVER RUNNING on {'http' if self.app.is_dev else 'sock'} ****************/")
        print(f"\
        \n\t- App env: {'DEV' if self.app.is_dev else 'PROD'}:{self.app.VERSION}\
        \n\t- App name: {self.app.name}\
        \n\t- App domain: {self.app.domain_name}\
        \n\t- App host: {self.app.host}\
        \n\t- App port: {self.app.port}\
        \n\t- App sock: {self.app.unix_socket_filepath}\
        \n\t- App ssl_key: {self.app.ssl_key_file}\
        \n\t- App ssl_cert: {self.app.ssl_cert_file}\
        \n\t- App Server: Uvicorn \
        \n\t- NGINX config: {self.app.nginx_config_filepath} \
        \n\t- System Version: {sys.version_info}")
        print(uvicorn_options)
        self.app.run_plugins(PyonirHooks.AFTER_INIT)
        # self.app.run_hooks(PyonirHooks.AFTER_INIT)
        self.is_active = True
        uvicorn.run(self.app.server, **uvicorn_options)

    def init_app_routes(self, routes: List[PyonirRoute], endpoint: str = ''):
        for path, func, methods, opts in routes:
            self.create_route(dec_func=func, path=f'{endpoint}{path}', methods=methods, **opts)
            pass

    def init_app_router(self, router: PyonirRouters):
        for endpoint, routes in router:
            self.init_app_routes(routes, endpoint=endpoint)

    @staticmethod
    def init_pyonir_endpoints(app: BaseApp):
        from pyonir.tests.backend.demo_controller import pyonir_ws_handler, pyonir_index
        # app.server.create_route(pyonir_ws_handler, "/sysws", ws=True)
        # app.server.create_route(pyonir_index, "/", methods='*')
        app.server.create_route(pyonir_index, "/{path:path}", methods='*')

    @staticmethod
    def create_static_route(assets_dir: str):
        """Creates a route for serving static files"""
        from starlette.staticfiles import StaticFiles
        return StaticFiles(directory=assets_dir)

    def create_route(self, dec_func: Optional[Callable],
                path: str = '',
                methods=None,
                static_path: str = None,
                auth: bool = None,
                ws: bool = None,
                sse: bool = None,
                **options: dict) -> Optional[Callable]:
        """A mapping of an HTTP method and an endpoint to a handler function."""
        import inspect

        is_async = inspect.iscoroutinefunction(dec_func) if dec_func else False
        is_asyncgen = inspect.isasyncgenfunction(dec_func) if dec_func else False
        if methods == '*':
            methods = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']
        if methods is None:
            methods = ['GET']

        if static_path:
            static_route = self.create_static_route(static_path)
            self.mount(path, static_route)
            self.url_map[path.lstrip('/')
            .replace('/', '_')] = {'path': path, 'dir': static_path, 'exists': os.path.exists(static_path)}
            return None

        route_name = dec_func.__name__ if dec_func else None
        docs = dec_func.__doc__ if dec_func else None
        route_path = path.split('/{')[0]
        name = route_name
        endpoint_route = path.split('/', 1)[0]
        # req_models = Site.server.url_map.get(route_name, {}).get('models') or {}

        new_route = {
            "doc": docs,
            "endpoint": endpoint_route,
            # "params": get_type_hints(dec_func) if dec_func else None,
            "route": path,  # has regex pattern
            "path": route_path,
            "methods": methods,
            "name": name,
            "auth": auth or options.pop('@auth', None),
            "sse": sse,
            "ws": ws,
            "async": is_async,
            "async_gen": is_asyncgen,
            "func": dec_func
        }
        # Add route path into categories
        self.endpoints.add(f"{endpoint_route}{route_path}")
        self.url_map[name] = new_route
        self.route_map[path] = new_route
        if sse:
            self.sse_routes.append(f"{endpoint_route}{route_path}")
        if auth:
            self.auth_routes.append(f"{endpoint_route}{route_path}")
        if ws:
            self.ws_routes.append(f"{endpoint_route}{route_path}")
            return self.add_websocket_route(path, dec_func, dec_func.__name__)

        async def dec_wrapper(star_req):
            pyonir_request = BaseRequest(star_req, self.app)
            res = await self.build_response(pyonir_request, dec_func)
            # self.app.reload_module(dec_func)
            return res

        if dec_func:
            dec_wrapper.__name__ = dec_func.__name__
            dec_wrapper.__doc__ = dec_func.__doc__
        self.add_route(path, dec_wrapper, methods=methods)

    def initialize_starlette(self):
        """Setup Starlette web server"""
        super().__init__()
        from starlette_wtf import CSRFProtectMiddleware
        from starlette.middleware.sessions import SessionMiddleware
        from starlette.middleware.trustedhost import TrustedHostMiddleware
        # from starlette.middleware.gzip import GZipMiddleware

        self.add_middleware(TrustedHostMiddleware)
        # star_app.add_middleware(GZipMiddleware, minimum_size=500)
        self.add_middleware(SessionMiddleware,
                                https_only=False,
                                secret_key=self.app.SECRET_SAUCE,
                                session_cookie=self.app.SESSION_KEY,
                                same_site='lax'
                                )
        self.add_middleware(CSRFProtectMiddleware, csrf_secret=self.app.SECRET_SAUCE)

    def get_virtual(self, url: str, virtual_data: dict = None) -> Union[tuple[str, dict, dict], tuple[None, None, None]]:
        """Performs url pattern matching against virtual routes and returns vitual page data and new path parameter values."""
        for vurl, vdata in (virtual_data).items():
            has_match = self.matching_route(url, vurl)
            if has_match:
                return vurl, vdata, has_match
        return None, None, None

    @staticmethod
    async def build_response(pyonir_request: BaseRequest, dec_func: callable):
        import inspect
        from pyonir import Site
        from pyonir.core.auth import Auth
        from pyonir.core.mapper import cls_mapper, func_request_mapper

        default_system_router = dec_func.__name__ == 'pyonir_index' if dec_func else False
        Site.server.request = pyonir_request
        if not default_system_router and dec_func and Site.is_dev:
            dec_func = pyonir_request.app.reload_module(dec_func, reload=True)

        # Init Web response object
        pyonir_response = BaseRestResponse(status_code=pyonir_request.status_code)

        # Preprocess request form data
        await pyonir_request.process_request_data()

        # Init Auth handler
        pyonir_request.auth = Auth(pyonir_request, Site)

        # Update template globals for request
        Site.TemplateEnvironment.globals['request'] = pyonir_request

        # File processing for request
        pyonir_request.set_app_context()
        req_file = pyonir_request.resolve_request_to_file(pyonir_request.app_ctx_ref)
        pyonir_request.file = req_file
        # app_ctx.apply_virtual_routes(pyonir_request) # disabled virtual routes in favor of module
        resolver = pyonir_request.process_resolver(pyonir_request)
        custom_response_headers = get_attr(pyonir_request.file.data, '@response.headers', {})

        route_func = dec_func if not callable(resolver) else resolver

        # Check resolver route security
        security = pyonir_request.auth.security.check(pyonir_request.auth)
        if security and not security.accepted:
            return Site.server.serve_redirect(security.redirect_to or '/')

        # Get router endpoint from map
        is_async = inspect.iscoroutinefunction(route_func)
        args = func_request_mapper(route_func, pyonir_request)
        if not pyonir_request.is_static:
            if callable(route_func):
                pyonir_request.server_response = await route_func(**args) if is_async else route_func(**args)
            else:
                pyonir_request.server_response = req_file.resolver or req_file.data

        # Perform redirects
        if pyonir_request.redirect_to:
            return Site.server.serve_redirect(pyonir_request.redirect_to)

        # Derive status code
        is_router = Site.server.url_map.get(route_func.__name__ if route_func else '') and not default_system_router
        pyonir_request.derive_status_code(is_router_method=is_router)

        # Execute plugins hooks initial request
        await Site.run_async_plugins(PyonirHooks.ON_REQUEST, pyonir_request)

        # Finalize response output
        if isinstance(pyonir_request.server_response, BaseRestResponse):
            pyonir_response = pyonir_request.server_response
        elif pyonir_request.is_static:
            pyonir_request.server_response = Site.server.resolve_static(Site, pyonir_request) or (await route_func(**args) if is_async else route_func(**args))
            pyonir_response.set_file_response(pyonir_request.server_response)
        else:
            if pyonir_request.type == JSON_RES:
                pyonir_response.set_json(pyonir_request.server_response or pyonir_request.file.data)
            elif pyonir_request.type == TEXT_RES:
                pyonir_response.set_html(pyonir_request.file.output_html(pyonir_request))
            else:
                print(f'{pyonir_request.type} doesnt have a handler')
            pyonir_response.set_media(pyonir_request.type)

        # Set headers
        if custom_response_headers:
            pyonir_response.set_headers_from_dict(custom_response_headers)

        # Generate response
        return pyonir_response.build()

    @staticmethod
    def matching_route(route_path: str, regex_path: str, api_endpoint: str = '') -> Optional[dict]:
        """Returns path parameters when match is found for virtual routes"""
        from starlette.routing import compile_path
        path_regex, path_format, *args = compile_path(regex_path)
        match = path_regex.match(route_path)# check if request path matches the router regex
        trail_match = match or path_regex.match(route_path+'/')
        if trail_match:
            params = args[0] if args else {}
            res = trail_match.groupdict()
            for key, converter in params.items():
                res[key] = converter.convert(res[key])
            return res

    @staticmethod
    def resolve_static(app: BaseApp, request: BaseRequest) -> Union['FileResponse', 'PlainTextResponse']:
        from_frontend = request.path.startswith(app.frontend_assets_route)
        from_public = request.path.startswith(app.public_assets_route)
        base_path = app.frontend_assets_dirpath if from_frontend else app.public_assets_dirpath if from_public else ''
        req_path = request.parts[1:] if len(request.parts) > 1 else request.parts
        path = os.path.join(base_path, *req_path)
        has_path = os.path.exists(path)
        if not has_path: print('Pyonir: skipping static: '+ path)
        return BaseServer.serve_static(path) if has_path else None

    @staticmethod
    def serve_static(path: str) -> Optional['FileResponse']:
        from starlette.responses import FileResponse, PlainTextResponse
        has_path = os.path.exists(path)
        res = FileResponse(path, 200) if has_path else None
        if has_path and path.endswith('.js'):
            res.headers["Cross-Origin-Opener-Policy"] = "same-origin"
            res.headers["Cross-Origin-Embedder-Policy"] = "require-corp"
        return res

    @staticmethod
    def response_renderer(value, media_type):
        from starlette.responses import Response, StreamingResponse
        if media_type == EVENT_RES:
            return StreamingResponse(content=value, media_type=media_type)
        return Response(content=value, media_type=media_type)

    @staticmethod
    def serve_redirect(url: str, code=302):
        from starlette.responses import RedirectResponse
        res = RedirectResponse(url=url.strip(), status_code=code)
        res.headers["Cache-Control"] = "no-store, no-cache, must-revalidate, max-age=0"
        res.headers["Pragma"] = "no-cache"
        res.headers["Expires"] = "0"
        return res

    @staticmethod
    def generate_nginx_conf(app: BaseApp) -> bool:
        """Generates a NGINX conf file based on App configurations"""
        from pyonir.core.utils import get_attr, create_file
        nginx_app_baseurl = get_attr(app.env, "nginx.baseurl")
        nginx_conf = app.TemplateEnvironment.get_template("nginx.jinja.conf") \
            .render(
            app_name=app.name,
            app_name_id=app.name.replace(' ', '_').lower(),
            domain=app.domain_name,
            is_dev=app.is_dev,
            is_secure=app.is_secure,
            ssl_cert_file=app.ssl_cert_file,
            ssl_key_file=app.ssl_key_file,
            site_dirpath=app.app_dirpath,
            site_logs_dirpath=app.logs_dirpath,
            app_socket_filepath=app.unix_socket_filepath,
            app_ignore_logs=f"{app.PUBLIC_ASSETS_DIRNAME}|{app.UPLOADS_DIRNAME}|{app.FRONTEND_ASSETS_DIRNAME}",
            frontend_assets_route=app.frontend_assets_route,
            frontend_assets_dirpath=app.frontend_assets_dirpath,
            public_assets_route=app.public_assets_route,
            public_assets_dirpath=app.public_assets_dirpath,
            site_uploads_route=app.uploads_route,
            site_uploads_dirpath=app.uploads_dirpath,
            site_ssg_dirpath=app.ssg_dirpath,
            custom_nginx_locations=get_attr(app.server, 'nginx_locations')
        )

        return create_file(app.nginx_config_filepath, nginx_conf, False)

@dataclass
class BaseRestResponse:
    """Represents a REST response from the server."""

    status_code: int = 000
    """HTTP status code of the response, e.g., 200 for success, 404 for not found."""

    message: str = ''
    """Response message, typically a string describing the result of the request."""

    data: dict = field(default_factory=dict)
    """Response data, typically a dictionary containing the response payload."""

    _cookies: list = field(default_factory=list)
    _html: str = None
    _stream: any = None
    _file_response: any = None
    _media_type: str = None
    _server_response: object = None
    _headers: dict = field(default_factory=dict)

    @property
    def headers(self): return self._headers

    @property
    def content(self) -> 'Response':
        from starlette.responses import Response, StreamingResponse
        content = ''
        media_type = self._media_type
        if self._file_response:
            self._media_type = 'static'
            return self._file_response
        if self._stream:
            media_type = EVENT_RES
            content = StreamingResponse(content=self._stream, media_type=EVENT_RES)
        if self._html:
            media_type = TEXT_RES
            content = self._html
        if self.data:
            media_type = JSON_RES
            content = self.to_json()
        self._media_type = media_type
        return Response(content=content, media_type=media_type)

    def to_dict(self) -> dict:
        return {
            'status_code': self.status_code,
            'message': self.message,
            'data': self.data
        }

    def to_json(self) -> str:
        """Converts the response to a JSON serializable dictionary."""
        from pyonir.core.utils import json_serial
        import json
        return json.dumps(self.to_dict(), default=json_serial)

    def render(self):
        return self._server_response

    def set_header(self, key, value):
        """Sets header values"""
        self._headers[key] = value
        return self

    def set_json(self, value: dict):
        # json = request.file.output_json()
        self.data = value
        return self

    def set_html(self, value: str):
        """Sets the html response value"""
        self._html = value
        return self

    def set_file_response(self, value: any):
        """Sets the file response value"""
        self._file_response = value

    def build(self):
        """Builds the response object"""
        self.set_header('Server', 'Pyonir Web Framework')
        res = self.content
        if self.headers:
            for key, value in self.headers.items():
                res.headers[key] = str(value)

        return self.content

    def set_server_response(self):
        """Renders the starlette web response"""
        from starlette.responses import Response, StreamingResponse

        if self._media_type == EVENT_RES and self._stream:
            return StreamingResponse(content=self._stream, media_type=EVENT_RES)

        content = self._html if self._html else self.to_json()
        media_type = TEXT_RES if self._html else JSON_RES
        self._server_response = Response(content=content, media_type=media_type)
        if self._headers:
            self._server_response.headers.update(self._headers)

    def set_media(self, media_type: str):
        self._media_type = media_type

    def set_cookie(self, cookie: dict):
        """
        :param cookie:
            key="access_token"
            value=jwt_token,
            httponly=True,
            secure=True,
            samesite="strict",
            max_age=3600,
        :return:
        """
        self._cookies.append(cookie)

    def set_headers_from_dict(self, headers: dict):
        """Sets multiple header values from a dictionary"""
        if headers:
            self._headers.update(headers)






from memory_profiler import memory_usage
import json
import yaml
import toml
from pyonir.core.parser import DeserializeFile, serializer
from pyonir.core.parsely import process_lines
import time

data = {
    "name": "MyApp",
    "version": "1.0.0",
    "config": {
        "host": "localhost",
        "port": 8000,
        "debug": True,
        "database": {
            "url": "postgresql://localhost/db",
            "pool_size": 10
        }
    }
}

# PARSELY
data_str = DeserializeFile.loads(data)

# JSON
json_str = json.dumps(data)

# YAML
yaml_str = yaml.dump(data)

# TOML
toml_str = toml.dumps(data)

dlines = data_str.strip().splitlines()
COUNT = 10000

def parsely_loop():
    for _ in range(COUNT):
        process_lines(dlines, cursor=0, data_container={})

def toml_loop():
    for _ in range(COUNT):
        toml.load(yaml_str)

def yaml_loop():
    for _ in range(COUNT):
        yaml.safe_load(yaml_str)

def deser_loop():
    for _ in range(COUNT):
        DeserializeFile.load(data_str)

def json_loop():
    for _ in range(COUNT):
        json.loads(json_str)

def print_metrics(func):
    name = func.__name__
    start = time.time()
    mem_usage = memory_usage((func,))
    end = time.time()
    print(f"{name} execution time: {end - start:.4f} seconds")
    print(f"{name} peak memory usage: {max(mem_usage):.2f} MiB\n\n")

if __name__ == "__main__":
    print_metrics(parsely_loop)
    # print_metrics(deser_loop)
    print_metrics(json_loop)
    print_metrics(yaml_loop)



from pyonir.core.app import BaseApp, BasePlugin
from pyonir.core.database import BaseFSQuery
from pyonir.core.schemas import BaseSchema
from pyonir.core.server import BaseRequest, BaseServer


class PyonirApp(BaseApp):pass
class PyonirServer(BaseServer): pass
class PyonirRequest(BaseRequest): pass
class PyonirCollection(BaseFSQuery): pass
class PyonirSchema(BaseSchema): pass
class PyonirPlugin(BasePlugin): pass


from __future__ import annotations

import os, re
from dataclasses import dataclass
from typing import Optional, Union, Generator
from starlette.datastructures import UploadFile

from pyonir import PyonirRequest
from pyonir.core.database import BaseFSQuery

from enum import Enum, unique

@unique
class AudioFormat(Enum):
    MP3  = "mp3"
    WAV  = "wav"
    FLAC = "flac"
    AAC  = "aac"
    OGG  = "ogg"
    M4A  = "m4a"

    def __str__(self):
        return self.value

@unique
class VideoFormat(Enum):
    MP4  = "mp4"
    MKV  = "mkv"
    AVI  = "avi"
    MOV  = "mov"
    WEBM = "webm"
    FLV  = "flv"

    def __str__(self):
        return self.value

@unique
class ImageFormat(Enum):
    JPG   = "jpg"
    JPEG  = "jpeg"
    PNG   = "png"
    GIF   = "gif"
    BMP   = "bmp"
    TIFF  = "tiff"
    WEBP  = "webp"
    SVG   = "svg"

    def __str__(self):
        return self.value

def sanitize_filename(filename: str) -> str:
    """
    Sanitize a file name by removing spaces, extra dots, and unsafe characters.
    Keeps the file extension if present.

    Example:
        cmd: sanitize_filename("my file.name.txt")

        output: 'my_filename.txt'
    """
    # Split into name and extension
    name, ext = os.path.splitext(filename)

    # Replace spaces with underscores
    name = name.replace(" ", "_")

    # Remove dots and any characters not alphanumeric, underscore, or hyphen
    name = re.sub(r"[^A-Za-z0-9_-]", "", name)

    # Collapse multiple underscores
    name = re.sub(r"_+", "_", name).strip("_")

    return f"{name}{ext}"

def rotate_image_from_exif(image):
    from PIL import ExifTags
    try:
        # image = Image.open(image_path)
        # Get EXIF data
        exif = image.getexif()

        # Find the Orientation tag
        orientation_tag = None
        for tag_id, tag_name in ExifTags.TAGS.items():
            if tag_name == 'Orientation':
                orientation_tag = tag_id
                break

        if orientation_tag in exif:
            orientation = exif[orientation_tag]
            # Rotate the image based on EXIF orientation
            if orientation == 3:
                image = image.rotate(180, expand=True)
            elif orientation == 6:
                image = image.rotate(270, expand=True)
            elif orientation == 8:
                image = image.rotate(90, expand=True)

            # Remove the orientation tag to prevent double rotation by other viewers
            if orientation_tag in image.info:
                del image.info[orientation_tag]
            if "exif" in image.info:
                image.info["exif"] = None # Clear EXIF data related to orientation

        return image

    except Exception as e:
        print(f"Error processing image: {e}")
        return None


class BaseMedia:
    """Represents an image file and its details."""

    def __init__(self, path: str = None, app_ctx: 'AppCtx' = None):
        self._app_ctx = app_ctx
        self._thumbnails = {}
        self.file_path = path
        name, ext = os.path.splitext(os.path.basename(self.file_path))
        self.file_ext = ext
        filename_meta = self.decode_filename(self.file_name)
        self.data = filename_meta or self.get_media_data(self.file_path)
        self.has_encoded_filename = bool(filename_meta)
        pass

    @property
    def file_name(self):
        name, ext = os.path.splitext(os.path.basename(self.file_path))
        self.file_ext = ext
        return name

    @property
    def file_dirpath(self):
        return os.path.dirname(self.file_path)

    @property
    def file_dirname(self):
        return os.path.basename(os.path.dirname(self.file_path))

    @property
    def slug(self):
        return f"{self.file_dirname}/{self.file_name}{self.file_ext}"

    @property
    def url(self):
        return f"/{self.slug}"

    @property
    def is_thumb(self) -> bool:
        """Check if the file is a thumbnail based on naming convention."""
        return self.file_name.startswith(os.path.basename(self.file_dirpath))

    @property
    def thumbnails(self) -> dict:
        """Returns a dictionary all thumbnails for the image with width x height as key and url as value"""
        if not self._thumbnails:
            thumbs = self._get_all_thumbnails()
            for thumb in thumbs or []:
                width = thumb.data.get('width')
                height = thumb.data.get('height')
                self._thumbnails[f"{width}x{height}"] = thumb.url
        return self._thumbnails

    def resized_to(self, width, height) -> str:
        """Generates an image accordion to width and height parameters and returns url to the new resized image"""
        if not self._thumbnails.get(f'{width}x{height}'):
            thumbs = self.resize([(width, height)])
            self._thumbnails.update(thumbs)
        return self._thumbnails.get(f'{width}x{height}')

    def _get_all_thumbnails(self) -> Optional[Generator]:
        """Collects thumbnails for the image"""
        if self.is_thumb: return None
        from pyonir import Site
        from pyonir.core.database import query_fs
        app_ctx = Site.app_ctx if Site else self._app_ctx
        thumbs_dir = os.path.join(self.file_dirpath, self.file_name)
        files = query_fs(str(thumbs_dir), model=BaseMedia, app_ctx=app_ctx)
        return files

    def compress(self, quality: int = 85) -> None:
        """Compress image file in place."""
        self.compress_image(self.file_path, self.file_path, quality=quality)

    def rename_media_file(self, file_name: str = None) -> None:
        """Renames media file as b64 encoded value"""
        encoded_filename = self.encode_filename(self.file_path, self.data) if not self.has_encoded_filename else None
        if not (file_name or encoded_filename): return
        new_filepath = self.file_path.replace(self.file_name+self.file_ext, file_name or encoded_filename)
        os.rename(self.file_path, new_filepath)
        self.has_encoded_filename = new_filepath.endswith(os.path.basename(self.file_path))
        self.file_path = new_filepath

    def resize(self, sizes: list[tuple] = None) -> dict:
        """
        Resize each image and save to the upload path in corresponding image size and paths
        This happens after full size images are saved to the filesystem
        :param sizes: list of (width, height) tuples
        """
        from PIL import Image
        from pyonir import Site
        from pathlib import Path

        raw_img = Image.open(self.file_path)
        thumb_dirname = Site.UPLOADS_THUMBNAIL_DIRNAME if Site else self.file_name
        sizes = [Site.THUMBNAIL_DEFAULT] if not sizes and Site else sizes
        base_dirpath = os.path.dirname(self.file_path)
        resized = {}
        if sizes is None:
            raise ValueError("Sizes must be provided if Site is not configured.")
        try:
            for dimensions in sizes:
                width, height = dimensions
                # self._sizes.append(dimensions)
                img = raw_img.resize((width, height), Image.Resampling.BICUBIC)
                file_name = f'{self.file_name}--{width}x{height}'
                img_dirpath = os.path.join(base_dirpath, thumb_dirname)
                Path(img_dirpath).mkdir(parents=True, exist_ok=True)
                filepath = os.path.join(img_dirpath, file_name + self.file_ext)
                if not os.path.exists(filepath):
                    img.save(filepath)
                    resized[f'{width}x{height}'] = BaseMedia(filepath)
            return resized
        except Exception as e:
            raise

    @staticmethod
    def decode_filename(encoded_filename: str) -> Optional[dict]:
        """ Reverse of encode_filename. """
        from pyonir.core.utils import parse_url_params
        import base64

        try:
            # restore padding
            padding = "=" * (-len(encoded_filename) % 4)
            encoded_filename = encoded_filename.replace("_", ".") + padding
            raw = base64.urlsafe_b64decode(encoded_filename.encode()).decode()
            parsed = parse_url_params(raw)
            return parsed
        except Exception as e:
            return None

    @staticmethod
    def encode_filename(file_path: str, meta_data: dict = None) -> str:
        """
        Build filename as url encoded string, then Base64 encode (URL-safe, no '.' in output).
        """
        from urllib.parse import urlencode
        from datetime import datetime
        import base64

        file_name, file_ext = os.path.splitext(os.path.basename(file_path))
        created_date = int(datetime.now().timestamp())
        raw = urlencode(meta_data) if meta_data else f'name={file_name}&ext={file_ext}&created_on={created_date}'
        # URL-safe base64 (no + or /), strip padding '='
        b64 = base64.urlsafe_b64encode(raw.encode()).decode().rstrip("=")
        return b64+file_ext

    @staticmethod
    def get_media_data(media_file_path: str):
        from pymediainfo import MediaInfo
        media_info = MediaInfo.parse(media_file_path)
        media_track_file = media_info.tracks.pop(0)
        created_on = media_track_file.file_creation_date
        for track in media_info.tracks:
            if track.track_type == "Image":
                return {
                    "name": media_track_file.file_name,
                    "created_on": created_on,
                    "width": track.width,
                    "height": track.height,
                    "size": media_track_file.file_size,
                }
            if track.track_type == "Audio":
                dur = track.duration / 1000 if track.duration else None # ms → seconds
                return {
                    "codec": track.codec,
                    "duration": dur,
                    "bit_rate": track.bit_rate,
                    "channels": track.channel_s,
                    "sampling_rate": track.sampling_rate,
                    "size": media_track_file.file_size,
                }
            if track.track_type == "Video":
                return {
                    "codec": track.codec,
                    "duration": track.duration / 1000 if track.duration else None,  # ms → seconds
                    "width": track.width,
                    "height": track.height,
                    "frame_rate": track.frame_rate,
                    "bit_rate": track.bit_rate,
                    "size": media_track_file.file_size,
                }

    @staticmethod
    def compress_image(input_path: str, output_path: str, quality: int = 85) -> None:
        """
        Auto-detect format (JPEG, PNG, WebP) and apply compression.
        """

        from PIL import Image
        media_type = BaseMedia.media_type(os.path.splitext(input_path)[1])
        if media_type != "image":
            return
        img = Image.open(input_path)
        fmt = img.format.upper()  # e.g. "JPEG", "PNG", "WEBP"
        img = rotate_image_from_exif(img)

        if fmt in ("JPEG", "JPG"):
            # JPEG: lossy compression
            img.save(output_path, format="JPEG", quality=quality, optimize=True)

        elif fmt == "PNG":
            # PNG: lossless, but can optimize
            img.save(output_path, format="PNG", optimize=True)

        elif fmt == "WEBP":
            # WebP: supports both lossy/lossless
            img.save(output_path, format="WEBP", quality=quality, method=6)

        else:
            # Default fallback → save in original format
            img.save(output_path, format=fmt)

        print(f"Compressed {input_path} ({fmt}) → {output_path}")

    @staticmethod
    def media_type(ext: str) -> str:
        """Return the media type based on file extension."""
        ext = ext.lstrip('.').lower()

        if ext in (f.value for f in AudioFormat):
            return "audio"
        elif ext in (f.value for f in VideoFormat):
            return "video"
        elif ext in (f.value for f in ImageFormat):
            return "image"
        return "document"

@dataclass
class UploadOptions:
    """Options for uploading media files."""
    directory_name: str = ''
    """Directory name to save the uploaded file."""
    file_name: str = ''
    """Strict file name for the uploaded file."""
    limit: int = 0
    """Maximum number of files to upload."""
    as_series: bool = False
    """If true, appends an index to the file name for multiple uploads."""
    compress_quality: int = 55
    """Quality for image compression (1-100)."""
    dimensions: tuple = (256, 256)

class MediaManager:
    """Manage audio, video, and image documents."""
    default_media_dirname = 'media' # general directory name for all media types

    def __init__(self, app: 'BaseApp'):
        self.app = app
        self.supported_formats = {ImageFormat.JPG, ImageFormat.PNG, VideoFormat.MP4, AudioFormat.MP3}
        self._storage_dirpath: str = os.path.join(app.contents_dirpath, self.default_media_dirname)
        """Location on fs to save file uploads"""

    @property
    def storage_dirpath(self) -> str: return self._storage_dirpath

    def is_supported(self, ext: str) -> bool:
        """Check if the media file has a supported format."""
        ext = ext.lstrip('.').lower()
        return ext in {fmt.value for fmt in self.supported_formats}

    def add_supported_format(self, fmt: Union[ImageFormat, AudioFormat, VideoFormat, None]):
        """Add a supported media format."""
        self.supported_formats.add(fmt)

    def set_storage_dirpath(self, storage_dirpath):
        self._storage_dirpath = storage_dirpath
        return self

    def close(self):
        """Closes any open connections by resetting storage path"""
        self._storage_dirpath = os.path.join(self.app.contents_dirpath, self.default_media_dirname)

    def get_media(self, file_id: str) -> BaseMedia:
        """Retrieves user paginated media files"""
        mpath = os.path.join(self.storage_dirpath, file_id)
        mfile = BaseMedia(mpath)
        return mfile

    def get_medias(self, file_type: str) -> list[BaseMedia]:
        """Retrieves user paginated media files"""
        files = BaseFSQuery(self.storage_dirpath, model=BaseMedia, force_all=True)
        return list(files)

    def delete_media_dir(self, dir_name: str) -> bool:
        """Delete all files in a directory. Returns True if deleted."""
        from pathlib import Path
        dir_path = os.path.join(self.storage_dirpath, dir_name)
        if not Path(dir_path).exists(): return False
        for file in Path(dir_path).glob("*"):
            if file.is_file():
                file.unlink()
        Path(dir_path).rmdir()
        return True

    def delete_media(self, media_id: str) -> bool:
        """Delete file by ID. Returns True if deleted."""
        from pathlib import Path
        path = os.path.join(self.storage_dirpath, media_id)
        if not Path(path).exists(): return False
        Path(path).unlink()
        return True

    # --- General Uploading ---
    async def upload(self, request: PyonirRequest, upload_options: UploadOptions = None) -> list[BaseMedia]:
        """Uploads a resource into specified directory
        :param upload_options: upload config options
        :param request: PyonirRequest instance
        """
        resource_files = []
        file_name = upload_options.file_name if upload_options else None
        limit = upload_options.limit if upload_options else None
        directory_name = upload_options.directory_name if upload_options else None
        as_series = upload_options.as_series if upload_options else False
        compress_quality = upload_options.compress_quality if upload_options else None
        for file in request.files:
            if not file.size: continue
            series_index = len(resource_files) + 1
            if limit and series_index > limit: break
            if file_name:
                file.filename = f"{file_name}_{series_index}" if as_series else file_name
            # directory_name = directory_name or BaseMedia.media_type(file.filename.split('.').pop())
            media_file: str = await self._upload_bytes(file, directory_name=directory_name)
            if media_file:
                file_media = BaseMedia(path=media_file)
                file_media.compress(quality=compress_quality)
                resource_files.append(file_media)
        return resource_files

    async def _upload_bytes(self, file: UploadFile, directory_name: str = None) -> Optional[str]:
        """
        Save an uploaded video file to disk and return its filename.
        or upload a video to Cloudflare R2 and return the object key.
        """
        from pathlib import Path
        filename = sanitize_filename(file.filename)
        if not filename: return None
        resource_id = [directory_name, filename] if directory_name else [filename]
        path = os.path.join(self.storage_dirpath, *resource_id)
        Path(os.path.dirname(path)).mkdir(parents=True, exist_ok=True)
        with open(path, "wb") as buffer:
            while chunk := await file.read(1024 * 1024):  # 1MB chunks
                buffer.write(chunk)
        return path

    @staticmethod
    def media_type(ext: str) -> Optional[str]:
        """Return the media type based on file extension."""
        ext = ext.lstrip('.').lower()

        if ext in (f.value for f in AudioFormat):
            return "audio"
        elif ext in (f.value for f in VideoFormat):
            return "video"
        elif ext in (f.value for f in ImageFormat):
            return "image"
        return None





from __future__ import annotations
from dataclasses import dataclass
from enum import StrEnum
from typing import Optional

from pyonir.core.schemas import BaseSchema
from pyonir.core.server import BaseRequest

# from pyonir.core import TaskAuthority

class PermissionLevel(str):
    NONE = 'none'
    """Defines the permission levels for users"""

    READ = 'read'
    """Permission to read data"""

    WRITE = 'write'
    """Permission to write data"""

    UPDATE = 'update'
    """Permission to update data"""

    DELETE = 'delete'
    """Permission to delete data"""

    ADMIN = 'admin'
    """Permission to perform administrative actions"""


@dataclass
class Role:
    """Defines the permissions for each role"""
    name: str
    perms: list[str]

    def to_dict(self, **kwargs) -> str:
        return self.name

    @classmethod
    def from_string(cls, role_name: str) -> "Role":
        """
        Create a Role instance from a string definition.

        Format: "RoleName:perm1,perm2,perm3"
        - RoleName is required.
        - Permissions are optional; defaults to [].

        Example:
            Role.from_string("Admin:read,write")
            -> Role(name="Admin", perms=["read", "write"])
        """
        role_name, perms = role_name.split(':')
        return cls(name=role_name.strip(), perms=perms.strip().split(',') if perms else [])


class Roles:
    """Defines the user roles and their permissions"""

    SUPER = Role(name='super', perms=[
        PermissionLevel.READ,
        PermissionLevel.WRITE,
        PermissionLevel.UPDATE,
        PermissionLevel.DELETE,
        PermissionLevel.ADMIN
    ])
    """Super user with all permissions"""
    ADMIN = Role(name='admin', perms=[
        PermissionLevel.READ,
        PermissionLevel.WRITE,
        PermissionLevel.UPDATE,
        PermissionLevel.DELETE
    ])
    """Admin user with most permissions"""
    AUTHOR = Role(name='author', perms=[
        PermissionLevel.READ,
        PermissionLevel.WRITE,
        PermissionLevel.UPDATE
    ])
    """Author user with permissions to create and edit content"""
    CONTRIBUTOR = Role(name='contributor', perms=[
        PermissionLevel.READ,
        PermissionLevel.WRITE
    ])
    """Contributor user with permissions to contribute content"""
    GUEST = Role(name='guest', perms=[
        PermissionLevel.READ
    ])
    """Contributor user with permissions to contribute content"""

    @classmethod
    def all_roles(cls):
        return [cls.SUPER, cls.ADMIN, cls.AUTHOR, cls.CONTRIBUTOR, cls.GUEST]

@dataclass
class UserSignIn(BaseSchema):
    """Represents a user sign in request"""

    email: str
    password: str

    def validate_email(self):
        """Validates the email format"""
        import re
        if not self.email:
            self._errors.append("Email cannot be empty")
        if not re.match(r"[^@]+@[^@]+\.[^@]+", self.email):
            self._errors.append(f"Invalid email address: {self.email}")

    def validate_password(self):
        """Validates the password for login"""
        if not self.password:
            self._errors.append("Password cannot be empty")
        elif len(self.password) < 6:
            self._errors.append("Password must be at least 6 characters long")

class UserMeta(BaseSchema):
    """Represents personal details about a user"""
    email: Optional[str] = ''
    first_name: Optional[str] = ''
    last_name: Optional[str] = ''
    gender: Optional[str] = ''
    age: Optional[int] = 0
    height: Optional[int] = 0
    weight: Optional[int] = 0
    phone: Optional[str] = ''
    about_you: Optional[str] = ''

class Location(BaseSchema):
    """Represents a user's location information."""
    ip: Optional[str]
    city: Optional[str]
    region: Optional[str]
    country_name: Optional[str]
    postal: Optional[str]
    latitude: Optional[float]
    longitude: Optional[float]
    signin_count: Optional[int] = 0
    device: Optional[str] = ''
    file_path: Optional[str] = ''
    file_dirpath: Optional[str] = ''

class User(BaseSchema):
    """Represents an app user"""
    _private_keys: list[str] = ['id', 'password', 'auth_token']
    """List of private keys that should not be included in JSON serialization"""

    # user signup fields
    # email: str
    """User's email address is required for signup"""
    password: str = ''
    """User's password to authenticate"""

    # configurable user details
    name: str = ''
    avatar: Optional[str] = ''
    meta: UserMeta = None

    # system specific fields
    uid: str = ''
    """Unique identifier for the user"""
    role: Role = Roles.GUEST
    """Role assigned to the user, defaults to 'none'"""
    verified_email: bool = False
    """Flag indicating if the user's email is verified"""
    file_path: str = ''
    """File path for user-specific files"""
    file_dirpath: str = ''
    """Directory path for user-specific files"""
    auth_from: Optional[str] = 'basic'
    """Authentication method used by the user (e.g., 'google', 'email')"""
    # signin_locations: Optional[list[Location]] = None
    # """Locations capture during signin"""
    auth_token: Optional[str] = ''
    """Authentication token verifying the user"""

    @property
    def email(self) -> str:
        return self.meta.email or ''

    @property
    def perms(self) -> list[PermissionLevel]:
        """Returns the permissions for the user based on their role"""
        user_role = getattr(Roles, self.role.name.upper())
        return user_role.perms if user_role else []

    def has_authority(self, authority: 'TaskAuthority') -> bool:
        """Checks if the user has a specific authority based on their role"""
        is_allowed = self.role in authority.roles
        return is_allowed

    def has_perm(self, action: PermissionLevel) -> bool:
        """Checks if the user has a specific permission based on their role"""
        user_role = getattr(Roles, self.role.name.upper(), Roles.GUEST)
        is_allowed = action in user_role.perms
        return is_allowed

    def has_perms(self, actions: list[PermissionLevel]) -> bool:
        return any([self.has_perm(action) for action in actions])

    def save_to_session(self, request: BaseRequest, key = None, value = None) -> None:
        """Convert instance to a serializable dict."""
        request.server_request.session[key or 'user'] = value or self.id

    @staticmethod
    def map_to_role(role_value: str) -> 'Role':
        """Maps a string role value to a Role instance"""
        r = getattr(Roles, role_value.upper(), None)
        return Role(name=role_value, perms=[]) if r is None else r


import os
from datetime import datetime
from typing import Optional, Type

from sqlmodel import Field

from pyonir import PyonirApp
from pyonir.core.schemas import BaseSchema
from pyonir.core.database import DatabaseService


class MockUser(BaseSchema, table_name='pyonir_users', primary_key='uid'):
    username: str
    email: str
    created_at: datetime = datetime.now
    uid: str = BaseSchema.generate_id


class MockDataService(DatabaseService):

    def create_table_from_model(self, model: Type[BaseSchema]) -> 'DatabaseService':
        pass

    name = "test_data_service"
    version = "0.1.0"
    endpoint = "/testdata"

    def create_table(self, sql_create: str) -> 'DatabaseService':
        return super().create_table(sql_create)

    def destroy(self):
        super().destroy()

    def connect(self) -> None:
        super().connect()

    def disconnect(self) -> None:
        super().disconnect()

    def insert(self, table: str, entity: MockUser) -> int:
        return super().insert(table, entity)

    def find(self, table: str, filter: dict = None) -> list:
        return super().find(table, filter)

    def update(self, table: str, id: int, data: dict) -> bool:
        if self.driver == "sqlite":
            pk = self.get_pk(table)
            set_clause = ', '.join(f"{k} = ?" for k in data.keys())
            query = f"UPDATE {table} SET {set_clause} WHERE {pk} = ?"
            values = list(data.values()) + [id]
            cursor = self.connection.cursor()
            cursor.execute(query, values)
            self.connection.commit()
            return cursor.rowcount > 0
        return False

    def delete(self, table: str, id: int) -> bool:
        if self.driver == "sqlite":
            pk = self.get_pk(table)
            cursor = self.connection.cursor()
            cursor.execute(f"DELETE FROM {table} WHERE {pk} = ?", (id,))
            self.connection.commit()
            return cursor.rowcount > 0
        return False


app = PyonirApp(__file__, False)  # Placeholder for PyonirApp instance
db = (MockDataService(app, "pyonir_test.db")
        .set_driver("sqlite").set_database())

def test_crud_operations():
    # Create
    db.connect()
    user = MockUser(username="testuser", email="test@example.com")
    table_name = user.__table_name__
    table_key = user.__primary_key__
    db.create_table(user._sql_create_table)
    user_id = db.insert(table_name, user)
    assert user_id

    # Read
    results = db.find(table_name, {table_key: user_id})
    assert (len(results) == 1)
    assert (results[0]["username"] == "testuser")
    assert (results[0]["email"] == "test@example.com")

    # Update
    updated = db.update(table_name, user_id, {
        "username": "newusername",
        "email": "newemail@example.com"
    })
    assert updated

    db.add_table_columns(table_name, {
        "age": "INTEGER DEFAULT 0"
    })

    # Verify update
    results = db.find(table_name, {table_key: user_id})
    assert (results[0]["username"] == "newusername")
    assert (results[0]["email"] == "newemail@example.com")
    assert (results[0]["age"] == 0)

    # Delete
    deleted = db.delete(table_name, user_id)
    assert (deleted)

    # Verify deletion
    results = db.find(table_name, {table_key: user_id})
    assert (len(results) == 0)

    db.disconnect()
    db.destroy()
    assert not os.path.exists(db.database)